#ifndef DATAREAD_H
#define DATAREAD_H

/*
  Copyright (C) 2001-2002 Leonardo Boselli

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the
    Free Software Foundation, Inc.,
    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  Please send remarks, questions and bug reports to
    boselli@uno.it
  or write to:
    Leonardo Boselli
    Via Paoletti, 33
    18100 Imperia
    ITALY
*/

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <math.h>

extern "C" {
#include "zlib\unzip.h"
#include "jpeg\jpeglib.h"
#include "png\png.h"
}

#include "exception.h"

#include "globject.h"
#include "fmsound.h"

class DRJpegReader;
class DRPngReader;
class DR3DSReader;
class DRResource;
class DRJpegResource;
class DRPngResource;
class DRSoundResource;
class DRMusicResource;
class DRFile;
class DRDataFile;
class DRZipFile;
class DRJpegFile;
class DRPngFile;
class DR3DSFile;
class DRData;
class DRImage;
class DRJpegImage;
class DRPngImage;

//
// DRJpegReader
//
class DRJpegReader {
protected:
  jpeg_decompress_struct cinfo;
  jpeg_error_mgr jerr;
public:
  DRJpegReader();
  static DRJpegImage* readImage(
    jpeg_decompress_struct* cinfo,
    jpeg_source_mgr* src,
    void (*init_source)(j_decompress_ptr cinfo),
    unsigned char (*fill_input_buffer)(j_decompress_ptr cinfo),
    void (*skip_input_data)(j_decompress_ptr cinfo,long),
    void (*term_source)(j_decompress_ptr cinfo),
    int alphaColor = -1, int tolerance = 0
  );
};

//
// DRPngReader
//
class DRPngReader {
private:
  LPBYTE readPointer;
protected:
  static void pngWarning(png_structp png_ptr, png_const_charp message);
  static void pngError(png_structp png_ptr, png_const_charp message);
  static void pngReadData(png_structp ptr, png_bytep data, png_size_t len);
public:
  DRPngReader();
  void setReadPointer(LPBYTE rp) {readPointer = rp;}
  LPBYTE getReadPointer() {return readPointer;}
  void moveReadPointer(int step) {readPointer += step;}
  DRPngImage* readImage() throw(Exception);
};

//
// DR3DSReader
//
class DR3DSReader {
private:
  HANDLE readFile;
protected:
  static void seekDataFromFile(DR3DSReader* reader, DWORD offset);
  static int dataLengthFromFile(DR3DSReader* reader);
  static DRImage* readImageFromFile(DR3DSReader* reader, char* imageName);
  static void readDataFromFile
    (DR3DSReader* reader, LPVOID buffer, DWORD bytesToBeRead, LPDWORD bytesRead);
private:
  DRZipFile* zipFile;
  LPBYTE bufferPointer;
  LPBYTE bufferStart;
  int bufferLength;
protected:
  static void seekDataFromZip(DR3DSReader* reader, DWORD offset);
  static int dataLengthFromZip(DR3DSReader* reader);
  static DRImage* readImageFromZip(DR3DSReader* reader, char* imageName);
  static void readDataFromZip
    (DR3DSReader* reader, LPVOID buffer, DWORD bytesToBeRead, LPDWORD bytesRead);
private:
  int chunkPointer;
  int chunkEndPointer;
  void (*seekDataPtr)(DR3DSReader* reader, DWORD offset);
  int (*dataLengthPtr)(DR3DSReader* reader);
  DRImage* (*readImagePtr)(DR3DSReader* reader, char* imageName);
  void (*readDataPtr)
    (DR3DSReader* reader, LPVOID buffer, DWORD bytesToBeRead, LPDWORD bytesRead);
protected:
  void seekData(DWORD offset) {seekDataPtr(this,offset);}
  int dataLength() {return dataLengthPtr(this);}
  DRImage* readImage(char* imageName) {return readImagePtr(this,imageName);}
  void readData(LPVOID buffer, DWORD bytesToBeRead, LPDWORD bytesRead)
    {readDataPtr(this,buffer,bytesToBeRead,bytesRead);}
protected:
  unsigned short getChunkID();
  int getChunkEndPointer() {return chunkEndPointer;}
  unsigned short readUShort();
  unsigned char readUChar();
  void read(char* string);
  void read(float* array, int arrayLength);
  void skip();
  void restore();
  bool isMoreDataAvailable(int endPointer) {return chunkPointer < endPointer;}
  enum tags {
    MAIN3DS       = 0x4d4d,
    EDIT3DS       = 0x3d3d,
    OBJECT3DS     = 0x4000,
    TRIMESH       = 0x4100,
    VERTEXLIST    = 0x4110,
    FACELIST      = 0x4120,
    MATERIALLIST  = 0x4130,
    MAPPINGLIST   = 0x4140,
    LOCALMATRIX   = 0x4160,
    MATERIALS     = 0xAFFF,
    MATERIALNAME  = 0xA000,
    MATAMBIENT    = 0xA010,
    MATDIFFUSE    = 0xA020,
    MATSPECULAR   = 0xA030,
    MATTEXTURE    = 0xA200,
    MATTEXNAME    = 0xA300,
    MATCOLOR      = 0x0011,
  };
public:
  DR3DSReader(DRZipFile* zip);
  DR3DSReader(HANDLE fileHandle);
  GLObjects* readMeshes(bool bumped = false);
};

//
// DRResource
//
class DRResource {
private:
  HRSRC resource;
protected:
  DRResource(int id) throw(Exception);
  DRResource(char* name) throw(Exception);
public:
  ~DRResource();
  LPBYTE getPointer() throw(Exception);
  DWORD getSize() {return SizeofResource(NULL,resource);}
};

//
// DRJpegResource
//
class DRJpegResource: public DRResource, protected DRJpegReader {
public:
  DRJpegResource(int id) throw(Exception);
  DRJpegImage* getImage(int alphaColor = -1, int tolerance = 0);
};

//
// DRPngResource
//
class DRPngResource: public DRResource, protected DRPngReader {
public:
  DRPngResource(int id) throw(Exception);
  DRPngImage* getImage() throw(Exception);
};

//
// DRSoundResource
//
class DRSoundResource: public DRResource {
public:
  DRSoundResource(int id) throw(Exception);
  FM3DSample* get3DSoundSample();
  FMSample* getSoundSample();
};

//
// DRMusicResource
//
class DRMusicResource: public DRResource {
public:
  DRMusicResource(int id) throw(Exception);
  FMMusic* getMusic();
};

//
// DRZipFile
//
class DRZipFile: protected DRJpegReader, protected DRPngReader {
private:
  unzFile fileHandler;
  unsigned char* zipBuffer;
  int zipBufferLength;
public:
  DRZipFile(char* zipFileName) throw(Exception);
  ~DRZipFile();
  unsigned char* getZipBuffer() {return zipBuffer;}
  int getZipBufferLength() {return zipBufferLength;}
  bool createZipBuffer(char* fileName);
  void deleteZipBuffer(bool deleteBuffer = true);
  bool findZippedFile(char* fileName)
    {return unzLocateFile(fileHandler,fileName,0) == UNZ_OK;}
  int getZippedFileSize() {
    unz_file_info ufi;
    unzGetCurrentFileInfo(fileHandler,&ufi,NULL,0,NULL,0,NULL,0);
    return ufi.uncompressed_size;
  }
  bool openZippedFile()
    {return unzOpenCurrentFile(fileHandler) == UNZ_OK;}
  bool closeZippedFile()
    {return unzCloseCurrentFile(fileHandler) != UNZ_CRCERROR;}
  int readZippedFile()
    {return unzReadCurrentFile(fileHandler,zipBuffer,zipBufferLength);}
  DRData* getData(char* fileName);
  DRImage* getImage(char* fileName);
  DRJpegImage* getJpegImage(char* fileName, int alphaColor = -1, int tol = 0);
  DRPngImage* getPngImage(char* fileName);
  GLObjects* getMeshes(char* fileName, bool bumped = false);
  FM3DSample* get3DSoundSample(char* fileName);
  FMSample* getSoundSample(char* fileName);
  FMMusic* getMusic(char* fileName);
};

//
// DRFile
//
class DRFile {
protected:
  char* fileName;
  HANDLE file;
public:
  DRFile(char* filNam) throw(Exception);
  ~DRFile();
  bool openFile();
  void closeFile();
  HANDLE getFileHandle() {return file;}
};

//
// DRDataFile
//
class DRDataFile: public DRFile {
public:
  DRDataFile(char* filNam) throw(Exception);
  DRData* getData();
};

//
// DRJpegFile
//
class DRJpegFile: public DRFile, protected DRJpegReader {
public:
  DRJpegFile(char* filNam) throw(Exception);
  DRJpegImage* getImage(int alphaColor = -1, int tolerance = 0);
};

//
// DRPngFile
//
class DRPngFile: public DRDataFile, protected DRPngReader {
public:
  DRPngFile(char* filNam) throw(Exception);
  DRPngImage* getImage();
};

//
// DRData
//
class DRData {
private:
  int size;
  unsigned char* buffer;
  bool ownsBuffer;
public:
  DRData(int sz, unsigned char* buf, bool ob = true);
  ~DRData();
  int getSize() const {return size;}
  unsigned char* getBuffer() {return buffer;}
  void setOwnsBuffer(bool ob) {ownsBuffer = ob;}
  bool getOwnsBuffer() {return ownsBuffer;}
};

//
// DRImage
//
class DRImage {
private:
  bool hasAlpha;
  bool grayScale;
  bool ownsBuffer;
  unsigned char* buffer;
  int width;
  int height;
protected:
  DRImage(
    int w, int h, unsigned char* b, bool ha = false,
    bool gs = false, bool ob = true
  );
public:
  DRImage(bool ha = false, bool gs = false, bool ob = true);
  ~DRImage();
  void setBuffer(unsigned char* b)
    {if(buffer) delete buffer; buffer = b;}
  void setWidth(int w) {width = w;}
  void setHeight(int h) {height = h;}
  void setHasAlpha(bool ha) {hasAlpha = ha;}
  void setGrayScale(bool gs) {grayScale = gs;}
  bool getOwnsBuffer() {return ownsBuffer;}
  unsigned char* getBuffer() {return buffer;}
  int getWidth() const {return width;}
  int getHeight() const {return height;}
  bool getHasAlpha() {return hasAlpha;}
  bool isGrayScale() {return grayScale;}
  void addAlpha(DRImage& img);
  void convertGray2RGBA();
  void convertGray2111A();
};

//
// DRJpegImage
//
class DRJpegImage: public DRImage {
public:
  DRJpegImage(int alphaColor = -1, bool ownsBuffer = true);
};

//
// DRPngImage
//
class DRPngImage: public DRImage {
public:
  DRPngImage(bool ha = false, bool ds = false, bool ob = true);
};

//
// DR3DSFile
//
class DR3DSFile: public DRFile {
public:
  DR3DSFile(char* filNam);
  GLObjects* getMeshes(bool bumped = false);
};

#endif //DATAREAD_H
