#ifndef DATASETS_H
#define DATASETS_H

/*
  Copyright (C) 2001-2002 Leonardo Boselli

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the
    Free Software Foundation, Inc.,
    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  Please send remarks, questions and bug reports to
    boselli@uno.it
  or write to:
    Leonardo Boselli
    Via Paoletti, 33
    18100 Imperia
    ITALY
*/

//
// DSStaticArray
//
template <class T> class DSStaticArray {
protected:
  unsigned short increment;
  unsigned short capacity;
  unsigned short size;
  T* elements;
public:
  DSStaticArray();
  DSStaticArray(int cap);
  ~DSStaticArray();
  unsigned short getCapacity() {return capacity;}
  unsigned short getSize() {return size;}
  void incrementCapacity();
  T& getElement(int idx) {return elements[idx];}
  void addElement(T& val)
    {if(size == capacity) incrementCapacity(); elements[size] = val; size++;}
};

//
// DSArray
//
template <class T, bool ownsElements = true> class DSArray:
  public DSStaticArray<T*>
{
public:
  DSArray();
  DSArray(int cap);
  ~DSArray();
  void incrementCapacity();
  T* setElement(int idx, T* val)
    {T* e = elements[idx]; elements[idx] = val; return e;}
  void removeElement(int idx);
  bool removeElement(T* element);
  void deleteElements();
};

//
// DSSortableObject
//
class DSSortableObject {
public:
  virtual float getSortIndex() = 0;
};

//
// DSSortedArray
//
template <class T, bool ownsElements = true> class DSSortedArray:
  public DSArray<T,ownsElements>
{
protected:
  T* setElement(int idx, T* val)
    {return DSArray<T,ownsElements>::setElement(idx,val);}
  void addElement(T* val) {return DSArray<T,ownsElements>::addElement(val);}
public:
  DSSortedArray();
  DSSortedArray(int cap);
  DSSortableObject* setElement(int idx, DSSortableObject* val)
    {return DSArray<T,ownsElements>::setElement(idx,(T*)val);}
  void addElement(DSSortableObject* val)
    {return DSArray<T,ownsElements>::addElement((T*)val);}
  void quickSort(int low,int high);
};

//
// DSChainNode
//
class DSChainNode {
protected:
  DSChainNode* next;
public:
  DSChainNode();
  virtual ~DSChainNode();
  DSChainNode* getNextNode() {return next;}
  void setNextNode(DSChainNode* node) {next = node;}
};

//
// DSTwoWayChainNode
//
class DSTwoWayChainNode: virtual public DSChainNode {
  DSTwoWayChainNode* prev;
public:
  DSTwoWayChainNode();
  DSTwoWayChainNode* getPrevNode() {return prev;}
  DSTwoWayChainNode* getNextNode()
  	{return dynamic_cast<DSTwoWayChainNode*>(next);}
  void setPrevNode(DSTwoWayChainNode* node) {prev = node;}
};

//
// DSChain
//
class DSChain {
protected:
  unsigned short length;
  DSChainNode* first;
  DSChainNode* last;
  DSChainNode* pointer;
public:
  DSChain();
  ~DSChain();
  DSChainNode* getFirstNode() {return first;}
  DSChainNode* getThisNode() {return pointer;}
  DSChainNode* getLastNode() {return last;}
  unsigned short getLength() {return length;}
  DSChainNode* reset() {return pointer = first;}
  DSChainNode* resetTo(DSChainNode* node) {return pointer = node;}
  DSChainNode* nextNode();
  void addNodeFirst(DSChainNode* node);
  void addNodeLast(DSChainNode* node);
  void insertNodeBehind(DSChainNode* node);
  bool deleteNode(DSChainNode* node);
  void deleteAll();
};

//
// DSTwoWayChain
//
class DSTwoWayChain: virtual public DSChain {
public:
  DSTwoWayChainNode* resetToLast()
  	{return dynamic_cast<DSTwoWayChainNode*>(pointer = last);}
  DSTwoWayChainNode* nextNode()
	  {return dynamic_cast<DSTwoWayChainNode*>(DSChain::nextNode());}
  DSTwoWayChainNode* prevNode();
  void addNodeFirst(DSTwoWayChainNode* node);
  void addNodeLast(DSTwoWayChainNode* node);
  void insertNodeBefore(DSTwoWayChainNode* node);
  void insertNodeBehind(DSTwoWayChainNode* node);
  bool deleteNode(DSTwoWayChainNode* node);
};

//
// DSListNode
//
template <class T, bool ownsElements = true> class DSListNode:
  virtual protected DSChainNode
{
  T* element;
public:
  DSListNode(T* e);
  ~DSListNode();
  T* getElement() {return element;}
  void setElement(T* e) {element = e;}
  DSListNode<T,ownsElements>* getNextNode() {
    return dynamic_cast<DSListNode<T,ownsElements>*>(
      DSChainNode::getNextNode()
    );
  }
  void setNextNode(DSListNode<T,ownsElements>* node)
    {DSChainNode::setNextNode(node);}
};

//
// DSTwoWayListNode
//
template <class T, bool ownsElements = true> class DSTwoWayListNode:
	public DSListNode<T,ownsElements>, protected DSTwoWayChainNode
{
public:
  DSTwoWayListNode(T* e);
  DSTwoWayListNode<T,ownsElements>* getNextNode()	{
  	return (DSTwoWayListNode<T,ownsElements>*)
		DSListNode<T,ownsElements>::getNextNode();
  }
  DSTwoWayListNode<T,ownsElements>* getPrevNode() {
  	return (DSTwoWayListNode<T,ownsElements>*)DSTwoWayChainNode::getPrevNode();
  }
  void setPrevNode(DSTwoWayListNode<T,ownsElements>* node)
  	{DSTwoWayChainNode::setPrevNode(node);}
};

//
// DSList
//
template <class T, bool ownsElements = true> class DSList:
  virtual protected DSChain
{
public:
  unsigned short getLength() {return DSChain::getLength();}
  DSListNode<T,ownsElements>* reset()
    {return dynamic_cast<DSListNode<T,ownsElements>*>(DSChain::reset());}
  DSListNode<T,ownsElements>* resetTo(DSListNode<T,ownsElements>* node)
    {return dynamic_cast<DSListNode<T,ownsElements>*>(DSChain::resetTo(node));}
  T* resetToFirstElement()
    {DSListNode<T,ownsElements>* n = reset(); return n? n->getElement(): 0;}
  T* resetToElementOf(DSListNode<T,ownsElements>* n)
    {resetTo(n); return n? n->getElement(): 0;}
  bool isEmpty() {return !getLength();}
  T* getFirstElement() {
    return
      isEmpty()?
      0:
      dynamic_cast<DSListNode<T,ownsElements>*>(getFirstNode())->getElement()
    ;
  }
  T* getLastElement() {
    return
      isEmpty()?
      0:
      dynamic_cast<DSListNode<T,ownsElements>*>(getLastNode())->getElement()
    ;
  }
  T* thisElement();
  T* nextElement();
  DSListNode<T,ownsElements>* nextNode()
    {return dynamic_cast<DSListNode<T,ownsElements>*>(DSChain::nextNode());}
  DSListNode<T,ownsElements>* addElement(T* e);
  DSListNode<T,ownsElements>* insertElement(T* e);
  bool deleteElement(T* e);
  bool removeElement(T* e);
  bool deleteNode(DSListNode<T,ownsElements>* n)
    {return DSChain::deleteNode((DSChainNode*)n);}
  bool removeNode(DSListNode<T,ownsElements>* n);
  void deleteAll() {DSChain::deleteAll();}
};

//
// DSTwoWayList
//
template <class T, bool ownsElements = true> class DSTwoWayList:
	public DSList<T,ownsElements>, protected DSTwoWayChain
{
public:
  DSTwoWayListNode<T,ownsElements>*
  reset() {
  	return dynamic_cast<DSTwoWayListNode<T,ownsElements>*>(
    	DSList<T,ownsElements>::reset()
    );
  }
  DSTwoWayListNode<T,ownsElements>*
  resetTo(DSTwoWayListNode<T,ownsElements>* node) {
  	return dynamic_cast<DSTwoWayListNode<T,ownsElements>*>(
    	DSList<T,ownsElements>::resetTo(node)
    );
  }
  DSTwoWayListNode<T,ownsElements>* resetToLast() {
  	return dynamic_cast<DSTwoWayListNode<T,ownsElements>*>(
    	DSTwoWayChain::resetToLast()
    );
  }
  T* resetToLastElement() {
  	DSTwoWayListNode<T,ownsElements>* n = resetToLast();
    return n? n->getElement(): NULL;
  }
  T* prevElement();
  DSTwoWayListNode<T,ownsElements>* nextNode()
  	{return (DSTwoWayListNode<T,ownsElements>*)DSTwoWayChain::nextNode();}
  DSTwoWayListNode<T,ownsElements>* prevNode()
  	{return (DSTwoWayListNode<T,ownsElements>*)DSTwoWayChain::prevNode();}
  DSTwoWayListNode<T,ownsElements>* addElement(T* e);
  DSTwoWayListNode<T,ownsElements>* insertElement(T* e);
  bool swapElements();
  bool deleteNode(DSTwoWayListNode<T,ownsElements>* n)
  	{return DSTwoWayChain::deleteNode(n);}
  bool removeNode(DSTwoWayListNode<T,ownsElements>* n);
};

#endif //DATASETS_H
