#ifndef FMSOUND_H
#define FMSOUND_H

/*
  Copyright (C) 2001-2002 Leonardo Boselli

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the
    Free Software Foundation, Inc.,
    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  Please send remarks, questions and bug reports to
    boselli@uno.it
  or write to:
    Leonardo Boselli
    Via Paoletti, 33
    18100 Imperia
    ITALY
*/

#define WIN32_LEAN_AND_MEAN
#include <windows.h>

#include "fmod\fmod.h"

class FMSoundManager;
class FMSample;
class FM3DSample;
class FMSound;
class FM3DSound;
class FM3DListener;
class FMMusic;

//
// FMSoundManager
//
class FMSoundManager {
private:
  static int channels;
public:
  enum {MIN_HARDWARE_CHANNELS = 16};
public:
  static void initialize();
  static void finalize() {FSOUND_Close();}
  static void update3DSound() {FSOUND_3D_Update();}
  static int getChannels() {return channels;}
};

//
// FMSample
//
class FMSample {
protected:
  FSOUND_SAMPLE* sample;
protected:
  FMSample();
  void init(int length, unsigned char* buffer, int flags);
public:
  FMSample(int length, unsigned char* buffer);
  ~FMSample();
  void setDefaults(int vol, int pan = -1, int freq = -1, int pri = -1)
    {FSOUND_Sample_SetDefaults(sample,freq,vol,pan,pri);}
  void setVolume(unsigned char vol) {setDefaults(-1,vol,-1,-1);}
  void setPan(unsigned char pan) {setDefaults(-1,-1,pan,-1);}
  void setFrequency(unsigned int freq) {setDefaults(freq,-1,-1,-1);}
  void setPriority(unsigned char pri) {setDefaults(-1,-1,-1,pri);}
  bool isLooping() {return FSOUND_Sample_GetMode(sample) & FSOUND_LOOP_NORMAL;}
  void setLooping(bool l)
    {FSOUND_Sample_SetLoopMode(sample, l? FSOUND_LOOP_NORMAL: FSOUND_LOOP_OFF);}
  FMSound* createSound();
  FSOUND_SAMPLE* getPointer() {return sample;}
};

//
// FM3DSample
//
class FM3DSample: public FMSample {
private:
  float invMinDistance;
public:
  FM3DSample(int length, unsigned char* buffer);
  FM3DSound* create3DSound();
  void setMinDistance(float min)
    {invMinDistance = 1/min; FSOUND_Sample_SetMinMaxDistance(sample,min,1e9f);}
  float getInvMinDistance() {return invMinDistance;}
};

//
// FMSound
//
class FMSound {
protected:
  int channel;
  FMSample* sample;
public:
  enum SoundStatus {
    IS_SOUND_PLAYING = 0,
    IS_CHANNEL_LOST = -1,
    WAS_SOUND_STOPPED = -2,
  };
public:
  FMSound(int c, FMSample* s);
  ~FMSound();
  bool isAudible() {return channel >= 0;}
  unsigned char getVolume() {
    int vol;
    if(isAudible()) return FSOUND_GetVolume(channel);
    else FSOUND_Sample_GetDefaults(sample->getPointer(),NULL,&vol,NULL,NULL);
    return vol;
  }
  void setVolume(unsigned char vol)
    {if(isAudible()) FSOUND_SetVolume(channel,vol);}
  void setPan(unsigned char pan)
    {if(isAudible()) FSOUND_SetPan(channel,pan);}
  void setFrequency(int freq)
    {if(isAudible()) FSOUND_SetFrequency(channel,freq);}
  void setPriority(unsigned char pri)
    {if(isAudible()) FSOUND_SetPriority(channel,pri);}
  void stop()
    {if(isAudible()) FSOUND_StopSound(channel); channel = WAS_SOUND_STOPPED;}
  void setMute(bool m)
    {if(isAudible()) FSOUND_SetMute(channel, m? TRUE: FALSE);}
  void play(unsigned char priority) {
    sample->setPriority(priority);
    channel = FSOUND_PlaySound(FSOUND_FREE,sample->getPointer());
  }
  SoundStatus getSoundStatus();
};

//
// FM3DSound
//
class FM3DSound: public FMSound {
public:
  FM3DSound(int c, FM3DSample* s);
  void setAttributes(float* pos = NULL, float* vel = NULL)
    {FSOUND_3D_SetAttributes(channel,pos,vel);}
  float getInvMinDistance() {return ((FM3DSample*)sample)->getInvMinDistance();}
};

//
// FM3DListener
//
class FMListener {
public:
  static void setAttributes(
    float fwdX, float fwdY, float fwdZ,
    float topX, float topY, float topZ,
    float* pos = NULL, float* vel = NULL
  ) {
    FSOUND_3D_Listener_SetAttributes(pos,vel,fwdX,fwdY,fwdZ,topX,topY,topZ);
  }
};

//
// FMMusic
//
class FMMusic {
private:
  FMUSIC_MODULE* module;
public:
  FMMusic(int length, unsigned char* buffer);
  ~FMMusic();
  void setVolume(unsigned char vol) {FMUSIC_SetMasterVolume(module,vol);}
  void play() {FMUSIC_PlaySong(module);}
  void stop() {FMUSIC_StopSong(module);}
  bool isPlaying() {return FMUSIC_IsPlaying(module) == TRUE;}
};

#endif // FMSOUND_H
