/*
  Copyright (C) 2001-2002 Leonardo Boselli

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the
    Free Software Foundation, Inc.,
    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  Please send remarks, questions and bug reports to
    boselli@uno.it
  or write to:
    Leonardo Boselli
    Via Paoletti, 33
    18100 Imperia
    ITALY
*/

#include "glbackground.h"

#include <stdlib.h> // for random()

//
// GLBackground
//
GLBackground::GLBackground(
  float r, float g, float b, float a, int src, int dst
):
  color(r,g,b,a), srcBlend(src), dstBlend(dst)
{}

GLBackground::~GLBackground() {
}

//
// GLStarfield
//

GLStarfield::GLStarfield(float rr, int num, GLTexture* txt, float sz):
  texture(txt->acquireReference())
{
  list = glGenLists(2);
  glNewList(list+0,GL_COMPILE);
  glPointSize(2);
  glBegin(GL_POINTS);
  generateStars(rr,num/4);
  glEnd();
  glPointSize(1.5f);
  glBegin(GL_POINTS);
  generateStars(rr,num/4);
  glEnd();
  glPointSize(1);
  glBegin(GL_POINTS);
  generateStars(rr,num/2);
  glEnd();
  glEndList();
  glNewList(list+1,GL_COMPILE);
  glBegin(GL_QUADS);
  for(int i = 0; i < 4; i++) {
    for(int j = 0; j < 4; j++) {
      float theta = 2*M_PI*(random(1000)-500)/1000.0;
      float phi = asin(2*(random(1000)-500)/1000.0);
      float rrcosphi = rr*cos(phi);
      float x = rrcosphi*cos(theta);
      float z = rrcosphi*sin(theta);
      float y = rr*sin(phi);
      M3Vector v1(x,y,z);
      M3Vector v2, v3;
      v1.plane(v2,v3);
      v2.scale(sz);
      v3.scale(sz);
      glTexCoord2f(0.25f*float(i),0.25f*float(j));
      glVertex3f(x,y,z);
      glTexCoord2f(0.25f*float(i+1),0.25f*float(j));
      glVertex3f(x+v2.getX(),y+v2.getY(),z+v2.getZ());
      glTexCoord2f(0.25f*float(i+1),0.25f*float(j+1));
      glVertex3f(
        x+v2.getX()+v3.getX(),y+v2.getY()+v3.getY(),z+v2.getZ()+v3.getZ()
      );
      glTexCoord2f(0.25f*float(i),0.25*float(j+1));
      glVertex3f(x+v3.getX(),y+v3.getY(),z+v3.getZ());
    }
  }
  glEnd();
  glEndList();
}

GLStarfield::~GLStarfield() {
  texture->releaseReference();
  glDeleteLists(list,2);
}

void GLStarfield::generateStars(float rr, int num) {
  for(int i = 0; i < num; i++) {
    float r = rr*pow(random(1000)/1000.0,0.33333f);
    float theta = 2*M_PI*(random(1000)-500)/1000.0;
    float phi = asin(2*(random(1000)-500)/1000.0);
    float rcosphi = r*cos(phi);
    float x = rcosphi*cos(theta);
    float z = rcosphi*sin(theta);
    float y = r*sin(phi);
    glVertex3f(x,y,z);
  }
}

void GLStarfield::render(bool blend) {
  if(!blend) glClear(GL_COLOR_BUFFER_BIT);
  glPushMatrix();
  glMultMatrixf(getTransform());
  bool isFoggy = glIsEnabled(GL_FOG);
  if(isFoggy) glDisable(GL_FOG);
  glColor3fv(getColor().getFloatArray());
  glDepthMask(GL_FALSE);
  glDisable(GL_LIGHTING);
  glEnable(GL_POINT_SMOOTH);
  glDisable(GL_TEXTURE_2D);
  glCallList(list+0);
  glEnable(GL_TEXTURE_2D);
  glEnable(GL_BLEND);
  glBlendFunc(GL_ONE,GL_ONE);
  glDisable(GL_POINT_SMOOTH);
  texture->apply();
  glCallList(list+1);
  glDisable(GL_BLEND);
  glEnable(GL_LIGHTING);
  glDepthMask(GL_TRUE);
  if(isFoggy) glEnable(GL_FOG);
  glPopMatrix();
}

//
// GLSkybox
//

GLSkybox::GLSkybox(int txtCount, GLTexture** txts, int src, int dst):
  GLBackground(1,1,1,1,src,dst)
{
  init(txtCount,txts);
}

void GLSkybox::init(int txtCount, GLTexture** txts) {
  texturesCount = txtCount;
  textures = new GLTexture*[texturesCount];
  for(int ct = 0; ct < texturesCount; ct++)
    textures[ct] = txts[ct]->acquireReference();
  list = glGenLists(texturesCount);
}

GLSkybox::GLSkybox(GLTexture** txts) {
  init(6,txts);
  const int halfSize = 100;
  glNewList(list+0,GL_COMPILE); // Top
  glBegin(GL_QUADS);
    glTexCoord2f(0,0); glVertex3f(-halfSize, halfSize,-halfSize);
    glTexCoord2f(0,1); glVertex3f( halfSize, halfSize,-halfSize);
    glTexCoord2f(1,1); glVertex3f( halfSize, halfSize, halfSize);
    glTexCoord2f(1,0); glVertex3f(-halfSize, halfSize, halfSize);
  glEnd();
  glEndList();
  glNewList(list+1,GL_COMPILE); // Left
  glBegin(GL_QUADS);
    glTexCoord2f(1,0); glVertex3f( halfSize, halfSize, halfSize);
    glTexCoord2f(0,0); glVertex3f( halfSize, halfSize,-halfSize);
    glTexCoord2f(0,1); glVertex3f( halfSize,-halfSize,-halfSize);
    glTexCoord2f(1,1); glVertex3f( halfSize,-halfSize, halfSize);
  glEnd();
  glEndList();
  glNewList(list+2,GL_COMPILE); // Front
  glBegin(GL_QUADS);
    glTexCoord2f(0,0); glVertex3f( halfSize, halfSize, halfSize);
    glTexCoord2f(0,1); glVertex3f( halfSize,-halfSize, halfSize);
    glTexCoord2f(1,1); glVertex3f(-halfSize,-halfSize, halfSize);
    glTexCoord2f(1,0); glVertex3f(-halfSize, halfSize, halfSize);
  glEnd();
  glEndList();
  glNewList(list+3,GL_COMPILE); // Right
  glBegin(GL_QUADS);
    glTexCoord2f(0,0); glVertex3f(-halfSize, halfSize, halfSize);
    glTexCoord2f(0,1); glVertex3f(-halfSize,-halfSize, halfSize);
    glTexCoord2f(1,1); glVertex3f(-halfSize,-halfSize,-halfSize);
    glTexCoord2f(1,0); glVertex3f(-halfSize, halfSize,-halfSize);
  glEnd();
  glEndList();
  glNewList(list+4,GL_COMPILE); // Back
  glBegin(GL_QUADS);
    glTexCoord2f(1,0); glVertex3f( halfSize, halfSize,-halfSize);
    glTexCoord2f(0,0); glVertex3f(-halfSize, halfSize,-halfSize);
    glTexCoord2f(0,1); glVertex3f(-halfSize,-halfSize,-halfSize);
    glTexCoord2f(1,1); glVertex3f( halfSize,-halfSize,-halfSize);
  glEnd();
  glEndList();
  glNewList(list+5,GL_COMPILE); // Bottom
  glBegin(GL_QUADS);
    glTexCoord2f(1,0); glVertex3f( halfSize,-halfSize, halfSize);
    glTexCoord2f(0,0); glVertex3f( halfSize,-halfSize,-halfSize);
    glTexCoord2f(0,1); glVertex3f(-halfSize,-halfSize,-halfSize);
    glTexCoord2f(1,1); glVertex3f(-halfSize,-halfSize, halfSize);
  glEnd();
  glEndList();
}

GLSkybox::~GLSkybox() {
  for(int ct = 0; ct < texturesCount; ct++) textures[ct]->releaseReference();
  if(textures) delete textures;
  glDeleteLists(list,texturesCount);
}

void GLSkybox::render(bool blend) {
  glPushMatrix();
  glMultMatrixf(getTransform());
  glDisable(GL_DEPTH_TEST);
  glDisable(GL_LIGHTING);
  bool isFoggy = glIsEnabled(GL_FOG);
  if(isFoggy) glDisable(GL_FOG);
  if(blend) {
    glEnable(GL_BLEND);
    glBlendFunc(getSrcBlend(),getDstBlend());
  }
  glColor3fv(getColor().getFloatArray());
  for(int ct = 0; ct < texturesCount; ct++) {
    textures[ct]->apply();
    glCallList(list+ct);
  }
  if(blend)
    glDisable(GL_BLEND);
  if(isFoggy) glEnable(GL_FOG);
  glEnable(GL_LIGHTING);
  glEnable(GL_DEPTH_TEST);
  glPopMatrix();
}

//
// GLMirroredSkybox
//

GLMirroredSkybox::GLMirroredSkybox(GLTexture** txts, int src, int dst):
  GLSkybox(5,txts,src,dst)
{
  const int halfSize = 100;
  glNewList(list+0,GL_COMPILE); // Top & Bottom
  glBegin(GL_QUADS);
    glTexCoord2f(0,0); glVertex3f(-halfSize, halfSize,-halfSize); // Top
    glTexCoord2f(0,1); glVertex3f( halfSize, halfSize,-halfSize);
    glTexCoord2f(1,1); glVertex3f( halfSize, halfSize, halfSize);
    glTexCoord2f(1,0); glVertex3f(-halfSize, halfSize, halfSize);
    glTexCoord2f(0,0); glVertex3f(-halfSize,-halfSize,-halfSize); // Bottom
    glTexCoord2f(1,0); glVertex3f(-halfSize,-halfSize, halfSize);
    glTexCoord2f(1,1); glVertex3f( halfSize,-halfSize, halfSize);
    glTexCoord2f(0,1); glVertex3f( halfSize,-halfSize,-halfSize);
  glEnd();
  glEndList();
  glNewList(list+1,GL_COMPILE); // Left
  glBegin(GL_QUADS);
    glTexCoord2f(1,0); glVertex3f( halfSize, halfSize, halfSize); // Top
    glTexCoord2f(0,0); glVertex3f( halfSize, halfSize,-halfSize);
    glTexCoord2f(0,1); glVertex3f( halfSize, 0       ,-halfSize);
    glTexCoord2f(1,1); glVertex3f( halfSize, 0       , halfSize);
    glTexCoord2f(1,1); glVertex3f( halfSize, 0       , halfSize); // Bottom
    glTexCoord2f(0,1); glVertex3f( halfSize, 0       ,-halfSize);
    glTexCoord2f(0,0); glVertex3f( halfSize,-halfSize,-halfSize);
    glTexCoord2f(1,0); glVertex3f( halfSize,-halfSize, halfSize);
  glEnd();
  glEndList();
  glNewList(list+2,GL_COMPILE); // Front
  glBegin(GL_QUADS);
    glTexCoord2f(0,0); glVertex3f( halfSize, halfSize, halfSize); // Top
    glTexCoord2f(0,1); glVertex3f( halfSize, 0       , halfSize);
    glTexCoord2f(1,1); glVertex3f(-halfSize, 0       , halfSize);
    glTexCoord2f(1,0); glVertex3f(-halfSize, halfSize, halfSize);
    glTexCoord2f(0,1); glVertex3f( halfSize, 0       , halfSize); // Bottom
    glTexCoord2f(0,0); glVertex3f( halfSize,-halfSize, halfSize);
    glTexCoord2f(1,0); glVertex3f(-halfSize,-halfSize, halfSize);
    glTexCoord2f(1,1); glVertex3f(-halfSize, 0       , halfSize);
  glEnd();
  glEndList();
  glNewList(list+3,GL_COMPILE); // Right
  glBegin(GL_QUADS);
    glTexCoord2f(0,0); glVertex3f(-halfSize, halfSize, halfSize); // Top
    glTexCoord2f(0,1); glVertex3f(-halfSize, 0       , halfSize);
    glTexCoord2f(1,1); glVertex3f(-halfSize, 0       ,-halfSize);
    glTexCoord2f(1,0); glVertex3f(-halfSize, halfSize,-halfSize);
    glTexCoord2f(0,1); glVertex3f(-halfSize, 0       , halfSize); // Bottom
    glTexCoord2f(0,0); glVertex3f(-halfSize,-halfSize, halfSize);
    glTexCoord2f(1,0); glVertex3f(-halfSize,-halfSize,-halfSize);
    glTexCoord2f(1,1); glVertex3f(-halfSize, 0       ,-halfSize);
  glEnd();
  glEndList();
  glNewList(list+4,GL_COMPILE); // Back
  glBegin(GL_QUADS);
    glTexCoord2f(1,0); glVertex3f( halfSize, halfSize,-halfSize); // Top
    glTexCoord2f(0,0); glVertex3f(-halfSize, halfSize,-halfSize);
    glTexCoord2f(0,1); glVertex3f(-halfSize, 0       ,-halfSize);
    glTexCoord2f(1,1); glVertex3f( halfSize, 0       ,-halfSize);
    glTexCoord2f(1,1); glVertex3f( halfSize, 0       ,-halfSize); // Bottom
    glTexCoord2f(0,1); glVertex3f(-halfSize, 0       ,-halfSize);
    glTexCoord2f(0,0); glVertex3f(-halfSize,-halfSize,-halfSize);
    glTexCoord2f(1,0); glVertex3f( halfSize,-halfSize,-halfSize);
  glEnd();
  glEndList();
}
