#ifndef GLFONT_H
#define GLFONT_H

/*
  Copyright (C) 2001-2002 Leonardo Boselli

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the
    Free Software Foundation, Inc.,
    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  Please send remarks, questions and bug reports to
    boselli@uno.it
  or write to:
    Leonardo Boselli
    Via Paoletti, 33
    18100 Imperia
    ITALY
*/

#include "opengl.h"
#include "gltexture.h"
#include "glmaterial.h"

//
// GLFont
//
class GLFont {
private:
  unsigned int listBase;
  int height;
  int width;
  int spacing;
public:
  GLFont(int h, int w = 0, int s = 0);
  ~GLFont();
  unsigned int getListBase() {return listBase;}
  static int getListLength() {return 256;}
  void setHeight(int value) {height = value;}
  void setWidth(int value) {width = value;}
  void setSpacing(int spac) {spacing = spac;}
  int getHeight() {return height;}
  int getWidth() {return width;}
  int getSpacing() {return spacing;}
  virtual void initialize() {glListBase(getListBase());}
  void setColor(float r, float g, float b) {glColor3f(r,g,b);}
  void drawText(int length, const char* text)
    {glCallLists(length,GL_UNSIGNED_BYTE,text);}
  void drawText(int x, int y, int length, const char* text)
    {setPosition(x,y); drawText(length,text);}
  void drawText(int x, int y, int z, int length, const char* text)
    {setPosition(x,y,z); drawText(length,text);}
  void flushText() {glFlush();}
  virtual void setPosition(float x, float y, float z = 0) = 0;
  virtual bool isTextured() = 0;
};

//
// GLTexturedFont
//
class GLTexturedFont: public GLFont {
private:
  GLTexture* texture;
  int listOffset;
public:
  GLTexturedFont(int h, int w, int spac, GLTexture* txt);
  void initialize() {glListBase(getListBase()-listOffset);}
  void applyTexture() {texture->apply();}
  void setPosition(float x, float y, float z = 0)
    {glLoadIdentity(); glTranslatef(x,y,z);}
  bool isTextured() {return true;}
};

//
// GLBitmapFont
//
class GLBitmapFont: public GLFont {
public:
  GLBitmapFont(
    HDC hDC, char* face, int height, int width = 0,
    int weight = FW_NORMAL, bool italic = false
  );
  void setPosition(float x, float y, float z = 0) {glRasterPos2f(x,y);}
  bool isTextured() {return false;}
};

//
// GLOutlineFont
//
class GLOutlineFont: public GLFont, public Reference {
private:
  GLYPHMETRICSFLOAT glyphMetricsFloat[256];
protected:
  void destroyObject();
public:
  GLOutlineFont(
    HDC hDC, char* face, int h, int w = 0, float depth = 0.0f,
    int weight = FW_NORMAL, bool italic = false
  );
  float getBoxWidth(char c) {return glyphMetricsFloat[c].gmfBlackBoxX;}
  float getBoxHeight(char c) {return glyphMetricsFloat[c].gmfBlackBoxY;}
  float getCellWidth(char c) {return glyphMetricsFloat[c].gmfCellIncX;}
  float getCellHeight(char c) {return glyphMetricsFloat[c].gmfCellIncY;}
  void setPosition(float x, float y, float z = 0) {glTranslatef(x,y,z);}
  bool isTextured() {return false;}
};

//
// GLText
//
class GLText: public GLObject {
private:
  GLOutlineFont* font;
  GLMaterial* material;
  char* text;
  float textWidth;
protected:
  void computeMaxRadius();
public:
  GLText(char* t, GLOutlineFont* f, GLMaterial* m = NULL);
  ~GLText();
  GLMaterial& getMaterial() {return *material;}
  void setMaterial(GLMaterial* mat) {
    if(material) material->releaseReference();
    material = mat->acquireReference();
  }
  char* getText() {return text;}
  void setText(char* txt) {text = txt; computeMaxRadius();}
  float getTextWidth() {return textWidth;}
  void render(GLCamera& camera);
};

//
// GLParagraph
//
class GLParagraph: public GLObjects {
public:
  void addText(GLText* txt) {addObject(txt);}
};

#endif // GLFONT_H
