#ifndef GLMATERIAL_H
#define GLMATERIAL_H

/*
  Copyright (C) 2001-2002 Leonardo Boselli

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the
    Free Software Foundation, Inc.,
    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  Please send remarks, questions and bug reports to
    boselli@uno.it
  or write to:
    Leonardo Boselli
    Via Paoletti, 33
    18100 Imperia
    ITALY
*/

#include "glcolor.h"
#include "gltexture.h"
#include "reference.h"

//
// GLMaterial
//
class GLMaterial: public Reference {
private:
  GLColor ambient;
  GLColor diffuse;
  GLColor specular;
  GLColor emissive;
  float shininess;
  GLTexture* diffuseTexture;
  GLTexture* glossTexture;
  float environment;
  GLTexture* environmentTexture;
public:
  GLMaterial* acquireReference()
    {return dynamic_cast<GLMaterial*>(Reference::acquireReference());}
protected:
  virtual void destroyObject();
public:
  GLMaterial();
  void setAmbient(float r, float g, float b, float a = 1)
    {ambient.set(r,g,b,a);}
  void setDiffuse(float r, float g, float b, float a = 1)
    {diffuse.set(r,g,b,a);}
  void setSpecular(float r, float g, float b, float a = 1)
    {specular.set(r,g,b,a);}
  void setEmissive(float r, float g, float b, float a = 1)
    {emissive.set(r,g,b,a);}
  void setShininess(float s) {shininess = s;}
  GLColor& getAmbient() {return ambient;}
  GLColor& getDiffuse() {return diffuse;}
  GLColor& getSpecular() {return specular;}
  GLColor& getEmissive() {return emissive;}
  float getShininess() {return shininess;}
  void setTransparency(float a) {
    ambient.setAlpha(a); diffuse.setAlpha(a);
    specular.setAlpha(a); emissive.setAlpha(a);
  }
  bool hasDiffuse() {return diffuseTexture != NULL;}
  void setDiffuseTexture(GLTexture* txt) {
    if(diffuseTexture) diffuseTexture->releaseReference();
    diffuseTexture = txt? txt->acquireReference(): NULL;
  }
  bool hasGloss() {return glossTexture != NULL;}
  void setGlossTexture(GLTexture* txt) {
    if(glossTexture) glossTexture->releaseReference();
    glossTexture = txt? txt->acquireReference(): NULL;
  }
  bool hasEnvironment() {return environmentTexture && environment > 0;}
  void setEnvironment(float env) {environment = env;}
  void setEnvironmentTexture(GLTexture* txt, float env = 1.0f) {
    environment = env;
    if(environmentTexture) environmentTexture->releaseReference();
    environmentTexture = txt? txt->acquireReference(): NULL;
  }
  void applyColor();
  void applyDiffuse() {diffuseTexture->apply();}
  void applyEnvironment();
  void applyGloss() {glossTexture->apply();}
};

//
// GLBumpedMaterial
//
class GLBumpedMaterial: public GLMaterial {
private:
  float bump;
  GLBumpedTexture* bumpTexture;
public:
  GLBumpedMaterial* acquireReference()
    {return dynamic_cast<GLBumpedMaterial*>(Reference::acquireReference());}
protected:
  void destroyObject();
public:
  GLBumpedMaterial();
  void setBump(float b) {bump = b;}
  float getBump() {return bump;}
  bool hasBump() {return bumpTexture && bump != 0;}
  void setBumpTexture(GLBumpedTexture* txt, float bmp = 0.1f) {
    bump = bmp;
    if(bumpTexture) bumpTexture->releaseReference();
    bumpTexture = txt? txt->acquireReference(): NULL;
  }
  void applyBumpDirect() {bumpTexture->applyDirect();}
  void applyBumpInverse() {bumpTexture->applyInverse();}
};

#endif // GLMATERIAL_H