/*
  Copyright (C) 2001-2002 Leonardo Boselli

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the
    Free Software Foundation, Inc.,
    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  Please send remarks, questions and bug reports to
    boselli@uno.it
  or write to:
    Leonardo Boselli
    Via Paoletti, 33
    18100 Imperia
    ITALY
*/

#include "glextensions.h"
#include "glmesh.h"

//
// GLTriangles
//
GLTriangles::GLTriangles(int triCount, unsigned short* iIndexes):
  count(triCount)
{
  int arrayLength = triCount*3;
  indexes = new unsigned short[arrayLength];
  for(int ct = 0; ct < arrayLength; ct++) indexes[ct] = iIndexes[ct];
}

GLTriangles::~GLTriangles() {
  if(indexes) delete indexes;
}

//
// GLVertices
//
GLVertices::GLVertices(
  int vertCount, float* iCoords, float* iNormals, float* iMappings
):
  count(vertCount)
{
  int coordsArrayLength = vertCount*3;
  coords = new float[coordsArrayLength];
  for(int ct = 0; ct < coordsArrayLength; ct++) coords[ct] = iCoords[ct];
  normals = new float[coordsArrayLength];
  if(iNormals)
    for(int ct = 0; ct < coordsArrayLength; ct++) normals[ct] = iNormals[ct];
  int mappingsArrayLength = vertCount+vertCount;
  mappings = new float[mappingsArrayLength];
  for(int ct = 0; ct < mappingsArrayLength; ct++) mappings[ct] = iMappings[ct];
}

GLVertices::~GLVertices() {
  if(coords) delete coords;
  if(normals) delete normals;
  if(mappings) delete mappings;
}

//
// GLShape
//
GLShape::GLShape(
  int vertCount, float* vertCoord, float* vertNorm, float* vertMap,
  int triCount, unsigned short* triIndexes
):
  triangles(triCount,triIndexes),
  vertices(vertCount,vertCoord,vertNorm,vertMap),
  list(-1)
{
  if(vertNorm == NULL) {
    for(int ct = 0; ct < vertCount; ct++) {
      M3Vector normal;
      float x = vertices.getCoordX(ct);
      float y = vertices.getCoordY(ct);
      float z = vertices.getCoordZ(ct);
      for(int ctVert = 0; ctVert < vertCount; ctVert++) {
        if(
          (x == vertices.getCoordX(ctVert)) &&
          (y == vertices.getCoordY(ctVert)) &&
          (z == vertices.getCoordZ(ctVert))
        ) {
          for(int ctFace = 0; ctFace < triCount; ctFace++) {
            int a = triangles.getA(ctFace);
            int b = triangles.getB(ctFace);
            int c = triangles.getC(ctFace);
            if((ctVert == a) || (ctVert == b) || (ctVert == c)) {
              float baX = vertices.getCoordX(b)-vertices.getCoordX(a);
              float baY = vertices.getCoordY(b)-vertices.getCoordY(a);
              float baZ = vertices.getCoordZ(b)-vertices.getCoordZ(a);
              float caX = vertices.getCoordX(c)-vertices.getCoordX(a);
              float caY = vertices.getCoordY(c)-vertices.getCoordY(a);
              float caZ = vertices.getCoordZ(c)-vertices.getCoordZ(a);
              normal.add(baY*caZ-baZ*caY,baZ*caX-baX*caZ,baX*caY-baY*caX);
            }
          }
        }
      }
      normal.normalize();
      vertices.setNormalX(ct,normal.getX());
      vertices.setNormalY(ct,normal.getY());
      vertices.setNormalZ(ct,normal.getZ());
    }
  }
  compile();
}

void GLShape::destroyObject() {
  glDeleteLists(list,1);
  delete this;
}

void GLShape::compile() {
  if(list >= 0) glDeleteLists(list,1);
  list = glGenLists(1);
  glNewList(list,GL_COMPILE);
  glBegin(GL_TRIANGLES);
  float maxDist2 = 0;
  for(int triCt = 0; triCt < triangles.getCount(); triCt++) {
    float dist2, x, y, z;
    int index;
    index = triangles.getA(triCt);
    x = vertices.getCoordX(index);
    y = vertices.getCoordY(index);
    z = vertices.getCoordZ(index);
    dist2 = x*x+y*y+z*z;
    if(dist2 > maxDist2) maxDist2 = dist2;
    if(isMultiTexturingEnabled) {
      glMultiTexCoord2fvARB(GL_TEXTURE0_ARB,vertices.getMap(index));
      glMultiTexCoord2fvARB(GL_TEXTURE1_ARB,vertices.getMap(index));
    } else {
      glTexCoord2fv(vertices.getMap(index));
    }
    glNormal3fv(vertices.getNormal(index));
    glVertex3fv(vertices.getCoord(index));
    index = triangles.getB(triCt);
    x = vertices.getCoordX(index);
    y = vertices.getCoordY(index);
    z = vertices.getCoordZ(index);
    dist2 = x*x+y*y+z*z;
    if(dist2 > maxDist2) maxDist2 = dist2;
    if(isMultiTexturingEnabled) {
      glMultiTexCoord2fvARB(GL_TEXTURE0_ARB,vertices.getMap(index));
      glMultiTexCoord2fvARB(GL_TEXTURE1_ARB,vertices.getMap(index));
    } else {
      glTexCoord2fv(vertices.getMap(index));
    }
    glNormal3fv(vertices.getNormal(index));
    glVertex3fv(vertices.getCoord(index));
    index = triangles.getC(triCt);
    x = vertices.getCoordX(index);
    y = vertices.getCoordY(index);
    z = vertices.getCoordZ(index);
    dist2 = x*x+y*y+z*z;
    if(dist2 > maxDist2) maxDist2 = dist2;
    if(isMultiTexturingEnabled) {
      glMultiTexCoord2fvARB(GL_TEXTURE0_ARB,vertices.getMap(index));
      glMultiTexCoord2fvARB(GL_TEXTURE1_ARB,vertices.getMap(index));
    } else {
      glTexCoord2fv(vertices.getMap(index));
    }
    glNormal3fv(vertices.getNormal(index));
    glVertex3fv(vertices.getCoord(index));
  }
  maxRadius = sqrt(maxDist2);
  glEnd();
  glEndList();
}

//
// GLHalo
//
GLHalo::GLHalo(): color(), depth(-1) {
}

//
// GLMesh
//
GLMesh::GLMesh(GLShape* shp, GLMaterial* mat): shape(shp->acquireReference()) {
  if(!mat) mat = new GLMaterial();
  material = mat->acquireReference();
  setMaxRadius(shp->getMaxRadius());
}

GLMesh::~GLMesh() {
  if(shape) shape->releaseReference();
  if(material) material->releaseReference();
}

void GLMesh::render(GLCamera& camera) {
  glPushMatrix();
  glMultMatrixf(getTransform());
  if((!material->hasEnvironment()) || halo.isEnabled()) {
    if(isTransparent()) {
      glDisable(GL_CULL_FACE);
      glDepthMask(GL_FALSE);
      glEnable(GL_BLEND);
      glBlendFunc(GL_SRC_ALPHA,GL_ONE);
    }
    material->applyColor();
    if(material->hasDiffuse())
      material->applyDiffuse();
    else
      glDisable(GL_TEXTURE_2D);
    glCallList(shape->getDisplayList());
    if(!material->hasDiffuse())
      glEnable(GL_TEXTURE_2D);
    if(isTransparent()) {
      glEnable(GL_CULL_FACE);
      glDepthMask(GL_TRUE);
      glDisable(GL_BLEND);
    }
  }
  if((material->hasEnvironment()) && (!halo.isEnabled())) {
    glEnable(GL_BLEND);
    if(material->hasGloss()) {
      if(isMultiTexturingEnabled) {
        glDisable(GL_BLEND);
        glActiveTextureARB(GL_TEXTURE0_ARB);
        material->applyEnvironment();
        glTexEnvf(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_REPLACE);
        glEnable(GL_TEXTURE_GEN_S);
        glEnable(GL_TEXTURE_GEN_T);
        glActiveTextureARB(GL_TEXTURE1_ARB);
        glEnable(GL_TEXTURE_2D);
        material->applyGloss();
        glTexEnvf(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_MODULATE);
        glCallList(shape->getDisplayList());
        glActiveTextureARB(GL_TEXTURE1_ARB);
        glDisable(GL_TEXTURE_2D);
        glActiveTextureARB(GL_TEXTURE0_ARB);
        glDisable(GL_TEXTURE_GEN_S);
        glDisable(GL_TEXTURE_GEN_T);
        glEnable(GL_BLEND);
      } else {
        glDisable(GL_BLEND);
        glTexEnvf(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_REPLACE);
        glEnable(GL_TEXTURE_GEN_S);
        glEnable(GL_TEXTURE_GEN_T);
        material->applyEnvironment();
        glCallList(shape->getDisplayList());
        glDisable(GL_TEXTURE_GEN_S);
        glDisable(GL_TEXTURE_GEN_T);
        glEnable(GL_BLEND);
        glBlendFunc(GL_DST_COLOR,GL_ZERO);
        material->applyGloss();
        glCallList(shape->getDisplayList());
      }
      glTexEnvf(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_MODULATE);
      glBlendFunc(GL_ONE,GL_ONE_MINUS_SRC_COLOR);
      material->applyColor();
      if(material->hasDiffuse())
        material->applyDiffuse();
      else
        glDisable(GL_TEXTURE_2D);
      glCallList(shape->getDisplayList());
      if(!material->hasDiffuse())
        glEnable(GL_TEXTURE_2D);
      glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
    } else {
      material->applyColor();
      if(material->hasDiffuse())
        material->applyDiffuse();
      else
        glDisable(GL_TEXTURE_2D);
      glCallList(shape->getDisplayList());
      if(!material->hasDiffuse())
        glEnable(GL_TEXTURE_2D);
      glTexEnvf(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_REPLACE);
      glEnable(GL_TEXTURE_GEN_S);
      glEnable(GL_TEXTURE_GEN_T);
      material->applyEnvironment();
      glCallList(shape->getDisplayList());
      glDisable(GL_TEXTURE_GEN_S);
      glDisable(GL_TEXTURE_GEN_T);
      glTexEnvf(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_MODULATE);
    }
    glDisable(GL_BLEND);
  }
  if(halo.isEnabled()) {
    glEnable(GL_BLEND);
    glDepthMask(GL_FALSE);
    if(material->hasEnvironment()) {
      glEnable(GL_TEXTURE_GEN_S);
      glEnable(GL_TEXTURE_GEN_T);
      material->applyEnvironment();
    } else glDisable(GL_TEXTURE_2D);
    glColor4fv(halo.getColor().getFloatArray());
    GLTriangles& triangles = shape->getTriangles();
    GLVertices& vertices = shape->getVertices();
    for(int ct = 0; ct < triangles.getCount(); ct++) {
      glBegin(GL_TRIANGLES);
      int a = triangles.getA(ct);
      int b = triangles.getB(ct);
      int c = triangles.getC(ct);
      glNormal3fv(vertices.getNormal(a));
      glVertex3f(
        vertices.getCoordX(a)+halo.getDepth()*vertices.getNormalX(a),
        vertices.getCoordY(a)+halo.getDepth()*vertices.getNormalY(a),
        vertices.getCoordZ(a)+halo.getDepth()*vertices.getNormalZ(a)
      );
      glNormal3fv(vertices.getNormal(b));
      glVertex3f(
        vertices.getCoordX(b)+halo.getDepth()*vertices.getNormalX(b),
        vertices.getCoordY(b)+halo.getDepth()*vertices.getNormalY(b),
        vertices.getCoordZ(b)+halo.getDepth()*vertices.getNormalZ(b)
      );
      glNormal3fv(vertices.getNormal(c));
      glVertex3f(
        vertices.getCoordX(c)+halo.getDepth()*vertices.getNormalX(c),
        vertices.getCoordY(c)+halo.getDepth()*vertices.getNormalY(c),
        vertices.getCoordZ(c)+halo.getDepth()*vertices.getNormalZ(c)
      );
      glEnd();
    }
    if(material->hasEnvironment()) {
      glDisable(GL_TEXTURE_GEN_S);
      glDisable(GL_TEXTURE_GEN_T);
    }
    glDepthMask(GL_TRUE);
    glDisable(GL_BLEND);
  }
  glPopMatrix();
}

//
// GLBumpedMesh
//
GLBumpedMesh::GLBumpedMesh(GLShape* shp, GLBumpedMaterial* mat): GLMesh(shp,mat) {
  GLVertices& vertices = shp->getVertices();
  GLTriangles& triangles = shp->getTriangles();
  int vertCount = vertices.getCount();
  diffU = new float[vertCount];
  diffV = new float[vertCount];
  dirU = new M3Vector[vertCount];
  dirV = new M3Vector[vertCount];
  for(int vert = 0; vert < vertCount; vert++) {
    dirU[vert] = M3Vector();
    dirV[vert] = M3Vector();
    int triCount = triangles.getCount();
    for(int tri = 0; tri < triCount; tri++) {
      int idx1, idx2;
      if(triangles.getA(tri) == vert) {
        idx1 = triangles.getB(tri);
        idx2 = triangles.getC(tri);
      } else if(triangles.getB(tri) == vert) {
        idx1 = triangles.getC(tri);
        idx2 = triangles.getA(tri);
      } else if(triangles.getC(tri) == vert) {
        idx1 = triangles.getA(tri);
        idx2 = triangles.getB(tri);
      } else continue;
      M3Vector l1(vertices.getCoord(idx1));
      l1.subtract(vertices.getCoord(vert));
      l1.normalize();
      M3Vector l2(vertices.getCoord(idx2));
      l2.subtract(vertices.getCoord(vert));
      l2.normalize();
      float deltaU1 = vertices.getMapU(idx1)-vertices.getMapU(vert);
      float deltaU2 = vertices.getMapU(idx2)-vertices.getMapU(vert);
      float deltaV1 = vertices.getMapV(idx1)-vertices.getMapV(vert);
      float deltaV2 = vertices.getMapV(idx2)-vertices.getMapV(vert);
      M3Vector tmp(l1);
      tmp.scale(deltaU1);
      dirU[vert].add(tmp);
      tmp = l2;
      tmp.scale(deltaU2);
      dirU[vert].add(tmp);
      l1.scale(deltaV1);
      dirV[vert].add(l1);
      l2.scale(deltaV2);
      dirV[vert].add(l2);
    }
    dirU[vert].normalize();
    dirV[vert].normalize();
  }
}

GLBumpedMesh::~GLBumpedMesh() {
  if(diffU) delete diffU;
  if(diffV) delete diffV;
  if(dirU) delete dirU;
  if(dirV) delete dirV;
}

void GLBumpedMesh::bumpOffsets(GLCamera& camera) {
  const float MAX_EMBOSS = 0.05f;
  float bump = MAX_EMBOSS*getBumpedMaterial().getBump();
  GLVertices& vertices = shape->getVertices();
  if(camera.isNearestLightDirectional()) {
    M3Vector dirRelative(camera.getNearestLightVector());
    dirRelative.rotateT(getTransform());
    for(int ct = 0; ct < vertices.getCount(); ct++) {
      diffU[ct] = bump*dirRelative.dot(dirU[ct]);
      diffV[ct] = bump*dirRelative.dot(dirV[ct]);
    }
  } else {
    M3Vector lightPosRelative(camera.getNearestLightVector());
    lightPosRelative.subtract(
      getPositionX(),getPositionY(),getPositionZ()
    );
    lightPosRelative.rotateT(getTransform());
    for(int ct = 0; ct < vertices.getCount(); ct++) {
      M3Vector lpr(lightPosRelative);
      lpr.subtract(getVertices().getCoord(ct)).normalize();
      diffU[ct] = bump*lpr.dot(dirU[ct]);
      diffV[ct] = bump*lpr.dot(dirV[ct]);
    }
  }
}

void GLBumpedMesh::render(GLCamera& camera) {
  glPushMatrix();
  glMultMatrixf(getTransform());
  if(isTransparent()) {
    glDisable(GL_CULL_FACE);
    glDepthMask(GL_FALSE);
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA,GL_ONE);
  }
  GLTriangles& triangles = shape->getTriangles();
  GLVertices& vertices = shape->getVertices();
  if(getBumpedMaterial().hasBump()) {
    glEnable(GL_BLEND);
    if(isMultiTexturingEnabled) {
      glActiveTextureARB(GL_TEXTURE0_ARB);
      getBumpedMaterial().applyBumpDirect();
      glTexEnvf(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_REPLACE);
      glActiveTextureARB(GL_TEXTURE1_ARB);
      glEnable(GL_TEXTURE_2D);
      getBumpedMaterial().applyBumpInverse();
      glTexEnvf(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_ADD);
      glDisable(GL_BLEND);
      bumpOffsets(camera);
      for(int ct = 0; ct < triangles.getCount(); ct++) {
        glBegin(GL_TRIANGLES);
        int a = triangles.getA(ct);
        int b = triangles.getB(ct);
        int c = triangles.getC(ct);
        glNormal3fv(vertices.getNormal(a));
        glMultiTexCoord2fvARB(GL_TEXTURE0_ARB,vertices.getMap(a));
        glMultiTexCoord2fARB(GL_TEXTURE1_ARB,
          vertices.getMapU(a)+diffU[a],vertices.getMapV(a)+diffV[a]
        );
        glVertex3fv(vertices.getCoord(a));
        glNormal3fv(vertices.getNormal(b));
        glMultiTexCoord2fvARB(GL_TEXTURE0_ARB,vertices.getMap(b));
        glMultiTexCoord2fARB(GL_TEXTURE1_ARB,
          vertices.getMapU(b)+diffU[b],vertices.getMapV(b)+diffV[b]
        );
        glVertex3fv(vertices.getCoord(b));
        glNormal3fv(vertices.getNormal(c));
        glMultiTexCoord2fvARB(GL_TEXTURE0_ARB,vertices.getMap(c));
        glMultiTexCoord2fARB(GL_TEXTURE1_ARB,
          vertices.getMapU(c)+diffU[c],vertices.getMapV(c)+diffV[c]
        );
        glVertex3fv(vertices.getCoord(c));
        glEnd();
      }
      glActiveTextureARB(GL_TEXTURE1_ARB);
      glDisable(GL_TEXTURE_2D);
      glActiveTextureARB(GL_TEXTURE0_ARB);
      glEnable(GL_BLEND);
    } else {
      glDisable(GL_BLEND);
      getBumpedMaterial().applyBumpDirect();
      glTexEnvf(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_DECAL);
      glCallList(shape->getDisplayList());
      glEnable(GL_BLEND);
      glBlendFunc(GL_ONE,GL_ONE);
      getBumpedMaterial().applyBumpInverse();
      bumpOffsets(camera);
      for(int ct = 0; ct < triangles.getCount(); ct++) {
        glBegin(GL_TRIANGLES);
        int a = triangles.getA(ct);
        int b = triangles.getB(ct);
        int c = triangles.getC(ct);
        glNormal3fv(vertices.getNormal(a));
        glTexCoord2f(vertices.getMapU(a)+diffU[a],vertices.getMapV(a)+diffV[a]);
        glVertex3fv(vertices.getCoord(a));
        glNormal3fv(vertices.getNormal(b));
        glTexCoord2f(vertices.getMapU(b)+diffU[b],vertices.getMapV(b)+diffV[b]);
        glVertex3fv(vertices.getCoord(b));
        glNormal3fv(vertices.getNormal(c));
        glTexCoord2f(vertices.getMapU(c)+diffU[c],vertices.getMapV(c)+diffV[c]);
        glVertex3fv(vertices.getCoord(c));
        glEnd();
      }
    }
    glTexEnvf(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_MODULATE);
    glBlendFunc(GL_DST_COLOR,GL_SRC_COLOR);
    getMaterial().applyColor();
    if(material->hasDiffuse())
      material->applyDiffuse();
    else
      glDisable(GL_TEXTURE_2D);
    glCallList(shape->getDisplayList());
    if(!material->hasDiffuse())
      glEnable(GL_TEXTURE_2D);
    glDisable(GL_BLEND);
  } else if((!getMaterial().hasEnvironment()) || (halo.isEnabled())) {
    getMaterial().applyColor();
    if(material->hasDiffuse())
      material->applyDiffuse();
    else
      glDisable(GL_TEXTURE_2D);
    glCallList(shape->getDisplayList());
    if(!material->hasDiffuse())
      glEnable(GL_TEXTURE_2D);
  }
  if(isTransparent()) {
    glEnable(GL_CULL_FACE);
    glDepthMask(GL_TRUE);
    glDisable(GL_BLEND);
  }
  if((getMaterial().hasEnvironment()) && (!halo.isEnabled())) {
    glEnable(GL_BLEND);
    if(getBumpedMaterial().hasBump()) {
      if(!getMaterial().hasGloss()) {
        glTexEnvf(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_REPLACE);
        glEnable(GL_TEXTURE_GEN_S);
        glEnable(GL_TEXTURE_GEN_T);
        getMaterial().applyEnvironment();
        glCallList(shape->getDisplayList());
        glDisable(GL_TEXTURE_GEN_S);
        glDisable(GL_TEXTURE_GEN_T);
        glTexEnvf(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_MODULATE);
      } else if((isMultiTexturingEnabled)&&(getMaterial().hasGloss())) {
        glActiveTextureARB(GL_TEXTURE0_ARB);
        glEnable(GL_TEXTURE_GEN_S);
        glEnable(GL_TEXTURE_GEN_T);
        getMaterial().applyEnvironment();
        glTexEnvf(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_REPLACE);
        glActiveTextureARB(GL_TEXTURE1_ARB);
        glEnable(GL_TEXTURE_2D);
        getMaterial().applyGloss();
        glTexEnvf(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_MODULATE);
        glBlendFunc(GL_ONE_MINUS_DST_COLOR,GL_ONE);
        glCallList(shape->getDisplayList());
        glActiveTextureARB(GL_TEXTURE1_ARB);
        glDisable(GL_TEXTURE_2D);
        glActiveTextureARB(GL_TEXTURE0_ARB);
        glDisable(GL_TEXTURE_GEN_S);
        glDisable(GL_TEXTURE_GEN_T);
        glTexEnvf(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_MODULATE);
        glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
      }
    } else if(getMaterial().hasGloss()) {
      if(isMultiTexturingEnabled) {
        glDisable(GL_BLEND);
        glActiveTextureARB(GL_TEXTURE0_ARB);
        getMaterial().applyEnvironment();
        glTexEnvf(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_REPLACE);
        glEnable(GL_TEXTURE_GEN_S);
        glEnable(GL_TEXTURE_GEN_T);
        glActiveTextureARB(GL_TEXTURE1_ARB);
        glEnable(GL_TEXTURE_2D);
        getMaterial().applyGloss();
        glTexEnvf(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_MODULATE);
        glCallList(shape->getDisplayList());
        glActiveTextureARB(GL_TEXTURE1_ARB);
        glDisable(GL_TEXTURE_2D);
        glActiveTextureARB(GL_TEXTURE0_ARB);
        glDisable(GL_TEXTURE_GEN_S);
        glDisable(GL_TEXTURE_GEN_T);
        glEnable(GL_BLEND);
      } else {
        glDisable(GL_BLEND);
        glTexEnvf(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_REPLACE);
        glEnable(GL_TEXTURE_GEN_S);
        glEnable(GL_TEXTURE_GEN_T);
        getMaterial().applyEnvironment();
        glCallList(shape->getDisplayList());
        glDisable(GL_TEXTURE_GEN_S);
        glDisable(GL_TEXTURE_GEN_T);
        glEnable(GL_BLEND);
        glBlendFunc(GL_DST_COLOR,GL_ZERO);
        getMaterial().applyGloss();
        glCallList(shape->getDisplayList());
      }
      glTexEnvf(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_MODULATE);
      glBlendFunc(GL_ONE,GL_ONE_MINUS_SRC_COLOR);
      getMaterial().applyColor();
      if(material->hasDiffuse())
        material->applyDiffuse();
      else
        glDisable(GL_TEXTURE_2D);
      glCallList(shape->getDisplayList());
      if(!material->hasDiffuse())
        glEnable(GL_TEXTURE_2D);
      glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
    } else {
      getMaterial().applyColor();
      if(material->hasDiffuse())
        material->applyDiffuse();
      else
        glDisable(GL_TEXTURE_2D);
      glCallList(shape->getDisplayList());
      if(!material->hasDiffuse())
        glEnable(GL_TEXTURE_2D);
      glTexEnvf(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_REPLACE);
      glEnable(GL_TEXTURE_GEN_S);
      glEnable(GL_TEXTURE_GEN_T);
      getMaterial().applyEnvironment();
      glCallList(shape->getDisplayList());
      glDisable(GL_TEXTURE_GEN_S);
      glDisable(GL_TEXTURE_GEN_T);
      glTexEnvf(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_MODULATE);
    }
    glDisable(GL_BLEND);
  }
  if(halo.isEnabled()) {
    glEnable(GL_BLEND);
    glDepthMask(GL_FALSE);
    if(getMaterial().hasEnvironment()) {
      glEnable(GL_TEXTURE_GEN_S);
      glEnable(GL_TEXTURE_GEN_T);
      getMaterial().applyEnvironment();
    } else glDisable(GL_TEXTURE_2D);
    glColor4fv(halo.getColor().getFloatArray());
    for(int ct = 0; ct < triangles.getCount(); ct++) {
      glBegin(GL_TRIANGLES);
      int a = triangles.getA(ct);
      int b = triangles.getB(ct);
      int c = triangles.getC(ct);
      glNormal3fv(vertices.getNormal(a));
      glVertex3f(
        vertices.getCoordX(a)+halo.getDepth()*vertices.getNormalX(a),
        vertices.getCoordY(a)+halo.getDepth()*vertices.getNormalY(a),
        vertices.getCoordZ(a)+halo.getDepth()*vertices.getNormalZ(a)
      );
      glNormal3fv(vertices.getNormal(b));
      glVertex3f(
        vertices.getCoordX(b)+halo.getDepth()*vertices.getNormalX(b),
        vertices.getCoordY(b)+halo.getDepth()*vertices.getNormalY(b),
        vertices.getCoordZ(b)+halo.getDepth()*vertices.getNormalZ(b)
      );
      glNormal3fv(vertices.getNormal(c));
      glVertex3f(
        vertices.getCoordX(c)+halo.getDepth()*vertices.getNormalX(c),
        vertices.getCoordY(c)+halo.getDepth()*vertices.getNormalY(c),
        vertices.getCoordZ(c)+halo.getDepth()*vertices.getNormalZ(c)
      );
      glEnd();
    }
    if(getMaterial().hasEnvironment()) {
      glDisable(GL_TEXTURE_GEN_S);
      glDisable(GL_TEXTURE_GEN_T);
    }
    glDepthMask(GL_TRUE);
    glDisable(GL_BLEND);
  }
  glPopMatrix();
}

