#ifndef GLMESH_H
#define GLMESH_H

/*
  Copyright (C) 2001-2002 Leonardo Boselli

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the
    Free Software Foundation, Inc.,
    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  Please send remarks, questions and bug reports to
    boselli@uno.it
  or write to:
    Leonardo Boselli
    Via Paoletti, 33
    18100 Imperia
    ITALY
*/

#include "globject.h"
#include "glmaterial.h"

//
// GLTriangles
//
class GLTriangles {
private:
  int count;
  unsigned short* indexes;
public:
  GLTriangles(int triCount, unsigned short* idx);
  ~GLTriangles();
  int getCount() {return count;}
  unsigned short* getIndexes() {return indexes;} 
  int getA(int face) {return indexes[face*3+0];}
  int getB(int face) {return indexes[face*3+1];}
  int getC(int face) {return indexes[face*3+2];}
};

//
// GLVertices
//
class GLVertices {
private:
  int count;
  float* coords;
  float* normals;
  float* mappings;
public:
  GLVertices(int vertCount, float* coo, float* norm, float* map);
  ~GLVertices();
  int getCount() {return count;}
  float* getCoords() {return coords;}
  float* getCoord(int idx) {return &coords[idx*3];}
  float getCoordX(int idx) {return coords[idx*3+0];}
  float getCoordY(int idx) {return coords[idx*3+1];}
  float getCoordZ(int idx) {return coords[idx*3+2];}
  float* getNormals() {return normals;}
  float* getNormal(int idx) {return &normals[idx*3];}
  float getNormalX(int idx) {return normals[idx*3+0];}
  float getNormalY(int idx) {return normals[idx*3+1];}
  float getNormalZ(int idx) {return normals[idx*3+2];}
  void setNormalX(int idx, float f) {normals[idx*3+0] = f;}
  void setNormalY(int idx, float f) {normals[idx*3+1] = f;}
  void setNormalZ(int idx, float f) {normals[idx*3+2] = f;}
  float* getMappings() {return mappings;}
  float* getMap(int idx) {return &mappings[idx*2];}
  float getMapU(int idx) {return mappings[idx*2+0];}
  float getMapV(int idx) {return mappings[idx*2+1];}
};

//
// GLShape
//
class GLShape: public Reference {
private:
  unsigned int list;
  GLTriangles triangles;
  GLVertices vertices;
  float maxRadius;
public:
  GLShape* acquireReference()
    {return dynamic_cast<GLShape*>(Reference::acquireReference());}
protected:
  void destroyObject();
public:
  GLShape(
    int vertCount, float* vertCoo, float* vertNorm, float* vertMap,
    int triCount, unsigned short* triIndexes
  );
  GLVertices& getVertices() {return vertices;}
  GLTriangles& getTriangles() {return triangles;}
  unsigned int getDisplayList() {return list;}
  float getMaxRadius() {return maxRadius;}
  void compile();
};

//
// GLHalo
//
class GLHalo {
private:
  float depth;
  GLColor color;
public:
  GLHalo();
  float getDepth() {return depth;}
  GLColor& getColor() {return color;}
  bool isEnabled() {return depth > 0;}
  void hide() {depth = -1;}
  void show(GLColor& clr, float dpt) {color.set(clr); depth = dpt;}
};

//
// GLMesh
//
class GLMesh: public GLObject {
protected:
  GLMaterial* material;
  GLShape* shape;
  GLHalo halo;
public:
  GLMesh(GLShape* shp, GLMaterial* mat = NULL);
  ~GLMesh();
  GLShape& getShape() {return *shape;}
  GLVertices& getVertices() {return shape->getVertices();}
  GLTriangles& getTriangles() {return shape->getTriangles();}
  void hideHalo() {setTransparent(false); halo.hide();}
  void showHalo(GLColor& clr, float dpt)
    {setTransparent(true); halo.show(clr,dpt);}
  GLMaterial& getMaterial() {return *material;}
  void setMaterial(GLMaterial* mat) {
    if(material) material->releaseReference();
    material = mat->acquireReference();
  }
  void compile() {shape->compile(); setMaxRadius(shape->getMaxRadius());}
  virtual GLMesh* clone() {return new GLMesh(shape,material);}
  virtual void render(GLCamera& camera);
};

//
// GLBumpedMesh
//
class GLBumpedMesh: public GLMesh {
private:
  float* diffU;
  float* diffV;
  M3Vector* dirU;
  M3Vector* dirV;
protected:
  void bumpOffsets(GLCamera& camera);
  void setMaterial(GLMaterial* mat) {GLMesh::setMaterial(mat);}
public:
  GLBumpedMesh(GLShape* shp, GLBumpedMaterial* mat = NULL);
  ~GLBumpedMesh();
  GLBumpedMaterial& getBumpedMaterial()
    {return (GLBumpedMaterial&)(getMaterial());}
  void setMaterial(GLBumpedMaterial* mat) {GLMesh::setMaterial(mat);}
  GLBumpedMesh* clone()
    {return new GLBumpedMesh(&getShape(),&getBumpedMaterial());}
  void render(GLCamera& camera);
};

#endif // GLMESH_H
