/*
  Copyright (C) 2001-2002 Leonardo Boselli

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the
    Free Software Foundation, Inc.,
    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  Please send remarks, questions and bug reports to
    boselli@uno.it
  or write to:
    Leonardo Boselli
    Via Paoletti, 33
    18100 Imperia
    ITALY
*/

#include "glparticle.h"

#include "glwin.h"

#include <stdlib.h> // for random()

//
// GLParticle
//

GLParticle::GLParticle():
  startPosition(), startVelocity(), life(-1)
{}

void GLParticle::render(
  M3Vector& pos, M3Vector& p1, M3Vector& p2, GLColor& color
) {
  glColor4fv(color.getFloatArray());
  glBegin(GL_TRIANGLE_STRIP);
  glTexCoord2f(1,1);
  glVertex3f(pos.getX()+p1.getX(),pos.getY()+p1.getY(),pos.getZ()+p1.getZ());
  glTexCoord2f(0,1);
  glVertex3f(pos.getX()+p2.getX(),pos.getY()+p2.getY(),pos.getZ()+p2.getZ());
  glTexCoord2f(1,0);
  glVertex3f(pos.getX()-p2.getX(),pos.getY()-p2.getY(),pos.getZ()-p2.getZ());
  glTexCoord2f(0,0);
  glVertex3f(pos.getX()-p1.getX(),pos.getY()-p1.getY(),pos.getZ()-p1.getZ());
  glEnd();
}

//
// GLEmitter
//

GLEmitter::GLEmitter(
  int pCount, float mxLif, float maxRadius, bool updateVelocity
):
  particlesCount(pCount), particleVelocity(),
  emitterVelocity(updateVelocity? new M3Vector(): NULL),
  speedVar(0), startSize(0), stepSize(0),
  startColor(), stepColor(), startGravity(), stepGravity(),
  maxLife(mxLif), particleTexture(NULL),
  oneShot(false), hasAlpha(false),
  lastRenderingTime(GLWin::getElapsedTime())
{
  setTransparent(true);
  setMaxRadius(maxRadius);
  particles = new GLParticle*[pCount];
  for(int ct = 0; ct < pCount; ct++) particles[ct] = new GLParticle();
  reset();
}

GLEmitter::~GLEmitter() {
  if(particleTexture) particleTexture->releaseReference();
  for(int ct = 0; ct < particlesCount; ct++) delete particles[ct];
  delete particles;
}

void GLEmitter::resetParticle(int idx, float life, float invLastRenderingStep) {
  M3Vector startVel(particleVelocity);
  if(emitterVelocity) startVel.add(*emitterVelocity);
  startVel.rotate(getTransform());
  M3Vector startPos(lastRenderingPosition);
  startPos.subtract(getPositionX(),getPositionY(),getPositionZ());
  startPos.scale(life*invLastRenderingStep);
  startPos.add(getPositionX(),getPositionY(),getPositionZ());
  float theta = 2*M_PI*(random(200)-100)*0.005f;
  float phi = M_PI*(random(200)-100)*0.005f;
  float speedVarCosPhi = speedVar*cos(phi);
  startVel.add(
    speedVarCosPhi*cos(theta),speedVarCosPhi*sin(theta),speedVar*sin(phi)
  );
  particles[idx]->setLife(life);
  particles[idx]->setStartPosition(startPos);
  particles[idx]->setStartVelocity(startVel);
}

void GLEmitter::reset() {
  float emissionRate = maxLife/particlesCount;
  for(int ct = 0; ct < particlesCount; ct++)
    particles[ct]->setLife(-ct*emissionRate);
}

void GLEmitter::render(GLCamera& camera) {
  float elapsedTime = GLWin::getElapsedTime();
  float lastRenderingStep = elapsedTime-lastRenderingTime;
  float invLastRenderingStep = 1/lastRenderingStep;
  if(emitterVelocity) {
    (*emitterVelocity = getPosition()).subtract(
      lastRenderingPosition
    ).scale(invLastRenderingStep);
  }
  glPushMatrix();
  glDisable(GL_CULL_FACE);
  glDepthMask(GL_FALSE);
  glEnable(GL_BLEND);
  if(hasAlpha)
    glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
  else
    glBlendFunc(GL_ONE,GL_ONE);
  glDisable(GL_LIGHTING);
  particleTexture->apply();
  for(int ct = 0; ct < particlesCount; ct++) {
    if(particles[ct]->getLife() <= 0) {
      if(particles[ct]->incrementLife(lastRenderingStep) >= 0)
        resetParticle(ct,particles[ct]->getLife(),invLastRenderingStep);
    } else {
      particles[ct]->incrementLife(lastRenderingStep);
    }
    if(particles[ct]->getLife() >= maxLife) {
      if(!oneShot) {
        resetParticle(
          ct,(float)fmod(particles[ct]->getLife(),maxLife),invLastRenderingStep
        );
      } else {
        particles[ct]->setLife(maxLife);
      }
    } else if(particles[ct]->getLife() >= 0) {
      float life = particles[ct]->getLife();
      float partSize = stepSize*life+startSize;
      M3Vector p1(
        camera.getSideDirectionX(),
        camera.isRenderingReflections()?
          -camera.getSideDirectionY():
           camera.getSideDirectionY(),
        camera.getSideDirectionZ()
      );
      p1.scale(partSize);
      M3Vector p2(
        camera.getUpDirectionX(),
        camera.isRenderingReflections()?
          -camera.getUpDirectionY():
           camera.getUpDirectionY(),
        camera.getUpDirectionZ()
      );
      p2.scale(partSize);
      GLColor color(stepColor);
      color.scale(life).add(startColor);
      M3Vector pos(particles[ct]->getStartPosition());
      M3Vector contribute(particles[ct]->getStartVelocity());
      contribute.scale(life);
      pos.add(contribute);
      contribute.set(startGravity);
      float life2 = life*life;
      contribute.scale(life2*0.5);
      pos.add(contribute);
      contribute.set(stepGravity);
      contribute.scale(life2*life/6);
      pos.add(contribute);
      particles[ct]->render(pos,p1,p2,color);
    }
  }
  lastRenderingTime = elapsedTime;
  lastRenderingPosition = getPosition();
  glEnable(GL_LIGHTING);
  glDisable(GL_BLEND);
  glDepthMask(GL_TRUE);
  glEnable(GL_CULL_FACE);
  glPopMatrix();
}
