/*
  Copyright (C) 2001-2002 Leonardo Boselli

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the
    Free Software Foundation, Inc.,
    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  Please send remarks, questions and bug reports to
    boselli@uno.it
  or write to:
    Leonardo Boselli
    Via Paoletti, 33
    18100 Imperia
    ITALY
*/

#include "glrenderer.h"

#include <stdlib.h> // for randomize()
#include <math.h>
#include <float.h>

//
// globals (for extensions)
//
bool isMultiTexturingEnabled = false;
int multiTextureUnits = 0;
PFNGLMULTITEXCOORD1FARBPROC glMultiTexCoord1fARB = NULL;
PFNGLMULTITEXCOORD2FARBPROC glMultiTexCoord2fARB = NULL;
PFNGLMULTITEXCOORD2FVARBPROC glMultiTexCoord2fvARB = NULL;
PFNGLMULTITEXCOORD3FARBPROC glMultiTexCoord3fARB = NULL;
PFNGLMULTITEXCOORD4FARBPROC glMultiTexCoord4fARB = NULL;
PFNGLACTIVETEXTUREARBPROC glActiveTextureARB = NULL;
PFNGLCLIENTACTIVETEXTUREARBPROC glClientActiveTextureARB = NULL;
BOOL (APIENTRY *wglSwapIntervalEXT)(int) = NULL;

bool isSubString(char* src, const char* search) {
  int maxpos = strlen(search)-1;
  for(unsigned int ct = 0; ct < strlen(src); ct++) {
    if((ct == 0) || ((ct > 1) && (src[ct-1] == ' '))) {
      int pos = 0;
      while (src[ct] != ' ') {
        if(src[ct] == search[pos]) pos++;
        if((pos > maxpos) && (src[ct+1] == ' ')) return true;
        ct++;
      }
    }
  }
  return false;
}

//
// GLRenderer
//

GLRenderer::GLRenderer() {
  _control87(MCW_EM,MCW_EM);
  randomize();
}

GLRenderer::~GLRenderer() {
}

bool GLRenderer::initializeRenderer() {
  char* extensions;
  extensions = strdup((char*)glGetString(GL_EXTENSIONS));
  if(
    isSubString(extensions,"GL_ARB_multitexture") &&
    isSubString(extensions,"GL_EXT_texture_env_add")
  ) {
    glGetIntegerv(GL_MAX_TEXTURE_UNITS_ARB,&multiTextureUnits);
    glMultiTexCoord1fARB = (PFNGLMULTITEXCOORD1FARBPROC)
      wglGetProcAddress("glMultiTexCoord1fARB");
    glMultiTexCoord2fARB = (PFNGLMULTITEXCOORD2FARBPROC)
      wglGetProcAddress("glMultiTexCoord2fARB");
    glMultiTexCoord2fvARB = (PFNGLMULTITEXCOORD2FVARBPROC)
      wglGetProcAddress("glMultiTexCoord2fvARB");
    glMultiTexCoord3fARB = (PFNGLMULTITEXCOORD3FARBPROC)
      wglGetProcAddress("glMultiTexCoord3fARB");
    glMultiTexCoord4fARB = (PFNGLMULTITEXCOORD4FARBPROC)
      wglGetProcAddress("glMultiTexCoord4fARB");
    glActiveTextureARB = (PFNGLACTIVETEXTUREARBPROC)
      wglGetProcAddress("glActiveTextureARB");
    glClientActiveTextureARB = (PFNGLCLIENTACTIVETEXTUREARBPROC)
      wglGetProcAddress("glClientActiveTextureARB");
    isMultiTexturingEnabled = true;
  } else {
    isMultiTexturingEnabled = false;
  }
  wglSwapIntervalEXT =
    (BOOL(APIENTRY*)(int))wglGetProcAddress("wglSwapIntervalEXT");
  applyVSync();
  glShadeModel(GL_SMOOTH);
  glClearColor(0,0,0,1);
  glClearDepth(1);
  glClearStencil(128);
  glClear(GL_STENCIL_BUFFER_BIT);
  glEnable(GL_DEPTH_TEST);
  glEnable(GL_LIGHTING);
  glEnable(GL_CULL_FACE);
  glDisable(GL_BLEND);
  glEnable(GL_TEXTURE_2D);
  glDisable(GL_TEXTURE_GEN_S);
  glDisable(GL_TEXTURE_GEN_T);
  glDisable(GL_POLYGON_SMOOTH);
  glDisable(GL_LINE_SMOOTH);
  glDisable(GL_POINT_SMOOTH);
  glDisable(GL_COLOR_MATERIAL);
  glDepthFunc(GL_LEQUAL);
  glDepthMask(GL_TRUE);
  glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
  glAlphaFunc(GL_GREATER,0);
  glHint(GL_PERSPECTIVE_CORRECTION_HINT,GL_NICEST);
  glHint(GL_FOG_HINT,GL_NICEST);
  glHint(GL_POLYGON_SMOOTH_HINT,GL_NICEST);
  glHint(GL_LINE_SMOOTH_HINT,GL_NICEST);
  glHint(GL_POINT_SMOOTH_HINT,GL_NICEST);
  glCullFace(GL_BACK);
  glFrontFace(GL_CCW);
  glTexEnvf(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_MODULATE);
  glTexGeni(GL_S,GL_TEXTURE_GEN_MODE,GL_SPHERE_MAP);
  glTexGeni(GL_T,GL_TEXTURE_GEN_MODE,GL_SPHERE_MAP);
  glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER,0);
  glLightModeli(GL_LIGHT_MODEL_TWO_SIDE,GL_FALSE);
  glMatrixMode(GL_TEXTURE);
  glLoadIdentity();
  glMatrixMode(GL_MODELVIEW);
  camera = new GLCamera();  
  world = new GLWorld();
  return true;
}

void GLRenderer::applyVSync() {
  wglSwapIntervalEXT? wglSwapIntervalEXT(vSyncOn? 1: 0): vSyncOn = true;
}

void GLRenderer::finalizeRenderer() {
  if(world) {
    delete world;
    world = NULL;
  }
  if(camera) {
    delete camera;
    camera = NULL;
  }
}

void GLRenderer::executeRendering() {
  glClear(GL_DEPTH_BUFFER_BIT);
  world->render(getCamera());
  if(isTextOverlayDrawn()) {
    glDisable(GL_LIGHTING);
    glDisable(GL_DEPTH_TEST);
    glMatrixMode(GL_PROJECTION);
    glPushMatrix();
    glLoadIdentity();
    glOrtho(0,getWidth(),0,getHeight(),-1,1);
    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
    glLoadIdentity();
    drawTextOverlay();
    glPopMatrix();
    glMatrixMode(GL_PROJECTION);
    glPopMatrix();
    glMatrixMode(GL_MODELVIEW);
    glEnable(GL_DEPTH_TEST);
    glEnable(GL_LIGHTING);
  }
};

