/*
  Copyright (C) 2001-2002 Leonardo Boselli

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the
    Free Software Foundation, Inc.,
    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  Please send remarks, questions and bug reports to
    boselli@uno.it
  or write to:
    Leonardo Boselli
    Via Paoletti, 33
    18100 Imperia
    ITALY
*/

#include "gltexture.h"
#include "glwin.h"

//
// GLTexture
//

GLTexture::GLTexture(unsigned int id): textureID(id) {
}

GLTexture::GLTexture(DRImage& img, bool repeat) {
  int w = img.getWidth();
  int h = img.getHeight();
  glGenTextures(1,&textureID);
  glBindTexture(GL_TEXTURE_2D,textureID);
  glPixelStorei(GL_UNPACK_ALIGNMENT,1);
  if(repeat) {
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_REPEAT);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_REPEAT);
  } else {
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_CLAMP);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_CLAMP);
  }
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
  glTexParameteri(
    GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR_MIPMAP_NEAREST
  );
  if(img.isGrayScale()) img.convertGray2RGBA();
  gluBuild2DMipmaps(
    GL_TEXTURE_2D,img.getHasAlpha()?GL_RGBA8:GL_RGB8,w,h,
    img.getHasAlpha()?GL_RGBA:GL_RGB,GL_UNSIGNED_BYTE,img.getBuffer()
  );
  glFlush();
}

void GLTexture::destroyObject() {
  glDeleteTextures(1,&textureID);
  delete this;
}

//
// GLAnimatedTexture
//

GLAnimatedTexture::GLAnimatedTexture(
  int imgsCount, DRImage** imgs, float duration, bool repeat
):
  GLTexture((unsigned int)imgsCount), loopDuration(duration),
  textureIDs(new unsigned int[imgsCount]), animated(true)
{
  glGenTextures(imgsCount,textureIDs);
  for(int ct = 0; ct < imgsCount; ct++) {
    int w = imgs[ct]->getWidth();
    int h = imgs[ct]->getHeight();
    glBindTexture(GL_TEXTURE_2D,textureIDs[ct]);
    glPixelStorei(GL_UNPACK_ALIGNMENT,1);
    if(repeat) {
      glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_REPEAT);
      glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_REPEAT);
    } else {
      glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_CLAMP);
      glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_CLAMP);
    }
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
    glTexParameteri(
      GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR_MIPMAP_NEAREST
    );
    if(imgs[ct]->isGrayScale()) imgs[ct]->convertGray2RGBA();
    gluBuild2DMipmaps(
    GL_TEXTURE_2D,imgs[ct]->getHasAlpha()?GL_RGBA8:GL_RGB8,w,h,
      imgs[ct]->getHasAlpha()?GL_RGBA:GL_RGB,GL_UNSIGNED_BYTE,
      imgs[ct]->getBuffer()
    );
  }
  glFlush();
}

void GLAnimatedTexture::destroyObject() {
  glDeleteTextures(getTexturesCount(),textureIDs);
  delete textureIDs;
  delete this;
}

void GLAnimatedTexture::apply() {
  if(isAnimated()) {
    double integer;
    apply((int)
      (modf(GLWin::getElapsedTime()/loopDuration,&integer)*getTexturesCount())
    );
  } else {
    apply(0);
  }
}

//
// GLBumpedTexture
//

GLBumpedTexture::GLBumpedTexture(DRImage& img, bool repeat) {
  int w = img.getWidth();
  int h = img.getHeight();
  glGenTextures(2,textureID);
  glPixelStorei(GL_UNPACK_ALIGNMENT,1);
  if(repeat) {
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_REPEAT);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_REPEAT);
  } else {
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_CLAMP);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_CLAMP);
  }
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
  glTexParameteri(
    GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR_MIPMAP_NEAREST
  );
  glBindTexture(GL_TEXTURE_2D,textureID[0]);
  glPixelTransferf(GL_RED_SCALE,0.5f);
  glPixelTransferf(GL_GREEN_SCALE,0.5f);
  glPixelTransferf(GL_BLUE_SCALE,0.5f);
  if(img.isGrayScale()) img.convertGray2RGBA();
  gluBuild2DMipmaps(
    GL_TEXTURE_2D,img.getHasAlpha()?GL_RGBA8:GL_RGB8,w,h,
    img.getHasAlpha()?GL_RGBA:GL_RGB,GL_UNSIGNED_BYTE,img.getBuffer()
  );
  glBindTexture(GL_TEXTURE_2D,textureID[1]);
  glPixelTransferf(GL_RED_SCALE,-0.5f);
  glPixelTransferf(GL_GREEN_SCALE,-0.5f);
  glPixelTransferf(GL_BLUE_SCALE,-0.5f);
  glPixelTransferf(GL_RED_BIAS,0.5f);
  glPixelTransferf(GL_GREEN_BIAS,0.5f);
  glPixelTransferf(GL_BLUE_BIAS,0.5f);
  gluBuild2DMipmaps(
    GL_TEXTURE_2D,img.getHasAlpha()?GL_RGBA8:GL_RGB8,w,h,
    img.getHasAlpha()?GL_RGBA:GL_RGB,GL_UNSIGNED_BYTE,img.getBuffer()
  );
  glPixelTransferf(GL_RED_SCALE,1);
  glPixelTransferf(GL_GREEN_SCALE,1);
  glPixelTransferf(GL_BLUE_SCALE,1);
  glPixelTransferf(GL_RED_BIAS,0);
  glPixelTransferf(GL_GREEN_BIAS,0);
  glPixelTransferf(GL_BLUE_BIAS,0);
  glFlush();
}

void GLBumpedTexture::destroyObject() {
  glDeleteTextures(2,textureID);
}
