#ifndef GLWIN_H
#define GLWIN_H

/*
  Copyright (C) 2001-2002 Leonardo Boselli

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the
    Free Software Foundation, Inc.,
    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  Please send remarks, questions and bug reports to
    boselli@uno.it
  or write to:
    Leonardo Boselli
    Via Paoletti, 33
    18100 Imperia
    ITALY
*/

#include "exception.h"
#include "glrenderer.h"

//
// GLWin
//
class GLWin: public GLRenderer {
private:
  HDC hDC;
  HWND hWnd;
  HGLRC hRC;
  HINSTANCE hInstance;
public:
  HDC getHDC() {return hDC;}
  HINSTANCE getHInstance() {return hInstance;}
private:
  bool active;
  bool running;
protected:
  void setActive(bool value) {active = value;}
public:
  bool isActive() {return active;}
  bool isRunning() {return running;}
private:
  int mouseKeys;
  int mouseDX, mouseDY;
  int centerX, centerY;
  bool mouseAcquired;
  bool keys[256];
public:
  void keyDown(int key) {keys[key] = true;}
  void keyUp(int key) {keys[key] = false;}
  bool isKeyPressed(int key) {return keys[key];}
  int getMouseDX() {int dx = mouseDX; mouseDX = 0; return dx;}
  int getMouseDY() {int dy = mouseDY; mouseDY = 0; return dy;}
  bool isControlPressed() {return mouseKeys & MK_CONTROL;}
  bool isShiftPressed() {return mouseKeys & MK_SHIFT;}
  bool isLeftButtonPressed() {return mouseKeys & MK_LBUTTON;}
  bool isMiddleButtonPressed() {return mouseKeys & MK_MBUTTON;}
  bool isRightButtonPressed() {return mouseKeys & MK_RBUTTON;}
  bool isMouseAcquired() {return mouseAcquired;}
  void acquireMouse(bool b)
    {ShowCursor(!(mouseAcquired = b)); if(b) centerMouse();}
  void centerMouse() {
    POINT pt; pt.x = centerX; pt.y = centerY;
    MapWindowPoints(hWnd,NULL,&pt,1);
    SetCursorPos(pt.x,pt.y);
  }
private:
  bool fullscreen;
  int fullWidth;
  int fullHeight;
protected:
  void setFullscreen(bool value) {fullscreen = value;}
  void setFullWidth(int w) {setWidth(fullWidth = w);}
  void setFullHeight(int h) {setHeight(fullHeight = h);}
  void setFullSize(int w, int h) {setFullWidth(w); setFullHeight(h);}
public:
  bool isFullscreen() {return fullscreen;}
  int getFullWidth() {return fullWidth;}
  int getFullHeight() {return fullHeight;}
private:
  GLFont* textOverlayFont;
public:
  GLFont* getTextOverlayFont() {return textOverlayFont;}
  void setTextOverlayFont(GLFont* font)
    {if(textOverlayFont) delete textOverlayFont; textOverlayFont = font;}
public:
  enum HelpMode {HIDDEN_HELP, REDUCED_HELP, USER_HELP, FULL_HELP};
private:
  HelpMode helpMode;
  bool showFramerate;
  bool showConsole;
public:
  HelpMode getHelpMode() {return helpMode;}
  void setHelpMode(HelpMode hm) {helpMode = hm;}
  bool isFramerateShown() {return showFramerate;}
  void setShowFramerate(bool value) {showFramerate = value;}
protected:
  bool isConsoleShown() {return showConsole;}
  void setShowConsole(bool value) {showConsole = value;}
  void drawTextOverlay();
  bool isTextOverlayDrawn() {return textOverlayFont &&
    (isFramerateShown() || (getHelpMode() != HIDDEN_HELP) || isConsoleShown());
  }
  virtual char* getHelpLine(int line) {return NULL;}
private:
  static float timeStep;
  static float invTimeStep;
  static float elapsedTime;
  static float ticksPeriod;
public:
  static float getTimeStep() {return timeStep;}
  static float getInvTimeStep() {return invTimeStep;}
  static float getElapsedTime() {return elapsedTime;}
protected:
  bool createWindow();
  void destroyWindow();
  void resizeWindow(GLsizei w, GLsizei h);
protected:
  static int askQuestion(char* message)
    {return MessageBox(NULL,message,"Please, answer.",MB_YESNO|MB_ICONQUESTION);}
  static void showError(char* message)
    {MessageBox(NULL,message,"ERROR",MB_OK|MB_ICONERROR);}
  static void showMessage(char* message)
    {MessageBox(NULL,message,"INFO",MB_OK|MB_ICONINFORMATION);}
  void shutdown(char* message) {destroyWindow(); showError(message);}
public:
  friend BOOL CALLBACK ::InitDlgProc(HWND hw, UINT msg, WPARAM wp, LPARAM lp);
  friend LRESULT CALLBACK ::WndProc(HWND hw, UINT um, WPARAM wp, LPARAM lp);
  friend int WINAPI ::WinMain(HINSTANCE hi, HINSTANCE hpi, LPSTR cl, int cs);
protected:
  GLWin() throw(Exception);
  virtual ~GLWin();
  // instance of the derived class (not defined in this module)
  static GLWin* newInstance(char* initStr) throw(Exception);
  // All these methods must be implemented by the application to make the
  // engine work. You must derive a new class from the GLWin, an instance
  // of which you must get through newInstance().
  virtual char* getTitle() {return "";}
  virtual bool initializeScene() = 0; // scene init (called on win mode change)
  virtual void finalizeScene() = 0; // scene final (called on win mode change)
  virtual void updateScene() = 0; // scene update
};

#endif // GLWIN_H
