#ifndef PHSIMULATOR_H
#define PHSIMULATOR_H

/*
  Copyright (C) 2001-2002 Leonardo Boselli

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the
    Free Software Foundation, Inc.,
    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  Please send remarks, questions and bug reports to
    boselli@uno.it
  or write to:
    Leonardo Boselli
    Via Paoletti, 33
    18100 Imperia
    ITALY
*/

#include "math3d.h"
#include "datasets.h"

class PHSimulator;
class PHEnvironment;
class PHBody;

//
// PHSimulator
//
class PHSimulator {
private:
  enum {INTEGRATIONS_PER_SECOND = 1000};
  static const M3Float DELTA_T;
private:
  DSStaticArray<M3Vector> positions0;
  DSStaticArray<M3Vector> positions1;
  DSStaticArray<M3Vector> accelerations;
  DSStaticArray<M3Vector>* positions;
  DSStaticArray<M3Vector>* oldPositions;
  DSArray<PHBody> bodies;
protected:
  void integrate();
  void zeroAcceleration(int idx) {accelerations.getElement(idx).set();}
  bool hasAcceleration(int idx) {
    M3Vector& acc = accelerations.getElement(idx);
    return acc.getX() != 0 || acc.getY() != 0 || acc.getZ() != 0;
  }
public:
  PHSimulator();
  void addBody(PHBody* body) {bodies.addElement(body);}
  int addPosition() {
    positions0.addElement(M3Vector());
    positions1.addElement(M3Vector());
    accelerations.addElement(M3Vector());
    return positions0.getSize()-1;
  }
  int addOrientation() {return addPosition();}
  M3Vector& getPosition(int idx) {return positions->getElement(idx);}
  M3Vector& getOldPosition(int idx) {return oldPositions->getElement(idx);}
  M3Vector& getAcceleration(int idx) {return accelerations.getElement(idx);}
  M3Vector& getOrientation(int idx) {return getPosition(idx);}
  M3Vector& getOldOrientation(int idx) {return getOldPosition(idx);}
  M3Vector& getAngularAcceleration(int idx) {return getAcceleration(idx);}
  int getBodiesCount() {return bodies.getSize();}
  PHBody* getBody(int idx) {return bodies.getElement(idx);}
  void runTime(M3Float timeStep);
};

//
// PHEnvironment
//
class PHEnvironment {
};

//
// PHBody
//
class PHBody {
private:
  int positionIndex;
  int orientationIndex;
  M3Float mass;
  M3Vector momentOfInertia;
protected:
  PHSimulator& simulator;
  PHEnvironment& environment;
  M3Reference& reference;
protected:
  M3Vector lastStepOrientation;
public:
  PHBody(PHSimulator& sim, PHEnvironment& env, M3Reference& ref);
  void setPosition(M3Float x, M3Float y, M3Float z)
    {simulator.getPosition(positionIndex).set(x,y,z);}
  void setVelocity(M3Float vx, M3Float vy, M3Float vz) {
    M3Vector& oldPos = simulator.getOldPosition(positionIndex);
    (oldPos = simulator.getPosition(positionIndex)).subtract(vx,vy,vx);
  }
  void setOrientation(M3Float theta, M3Float phi, M3Float rho) {
    lastStepOrientation =
      simulator.getPosition(orientationIndex).set(theta,phi,rho);
  }
  void setAngularVelocity(M3Float thetaDot, M3Float phiDot, M3Float rhoDot) {
    M3Vector& oldPos = simulator.getOldPosition(orientationIndex);
    oldPos = simulator.getPosition(orientationIndex);
    oldPos.subtract(thetaDot,phiDot,rhoDot);
  }
  virtual M3Float getMass() {return mass;}
  void setMass(M3Float m) {mass = m;}
  virtual M3Vector& getMomentOfInertia() {return momentOfInertia;}
  void setMomentOfInertia(M3Float mx, M3Float my, M3Float mz)
    {momentOfInertia.set(mx,my,mz);}
  void applyForce(M3Vector f)
    {simulator.getAcceleration(positionIndex).add(f.scale(1/getMass()));}
  void applyTorque(M3Vector t) {
    M3Vector& moi = getMomentOfInertia();
    simulator.getAngularAcceleration(positionIndex).add(
      t.getX()/moi.getX(),t.getY()/moi.getY(),t.getZ()/moi.getZ()
    );
  }
  void updateReference();
};

#endif // PHSIMULATOR_H

